'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import warnings
from typing import Tuple, Union, List, Dict, Any

import deprecation
import pandas as pd
from pandas import DataFrame

from pm4py.objects.bpmn.obj import BPMN
from pm4py.objects.heuristics_net.obj import HeuristicsNet
from pm4py.objects.log.obj import EventLog
from pm4py.objects.log.obj import EventStream
from pm4py.objects.petri_net.obj import PetriNet, Marking
from pm4py.objects.process_tree.obj import ProcessTree
from pm4py.util.pandas_utils import check_is_pandas_dataframe, check_pandas_dataframe_columns
from pm4py.utils import get_properties, xes_constants
from pm4py.objects.ocel.obj import OCEL
from pm4py.util import constants


def discover_dfg(log: Union[EventLog, pd.DataFrame]) -> Tuple[dict, dict, dict]:
    """
    Discovers a DFG from a log

    Parameters
    --------------
    log
        Event log

    Returns
    --------------
    dfg
        DFG
    start_activities
        Start activities
    end_activities
        End activities
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)
        from pm4py.util import constants
        properties = get_properties(log)
        from pm4py.algo.discovery.dfg.adapters.pandas.df_statistics import get_dfg_graph
        activity_key = properties[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] if constants.PARAMETER_CONSTANT_ACTIVITY_KEY in properties else xes_constants.DEFAULT_NAME_KEY
        timestamp_key = properties[constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] if constants.PARAMETER_CONSTANT_TIMESTAMP_KEY in properties else xes_constants.DEFAULT_TIMESTAMP_KEY
        case_id_key = properties[constants.PARAMETER_CONSTANT_CASEID_KEY] if constants.PARAMETER_CONSTANT_CASEID_KEY in properties else constants.CASE_CONCEPT_NAME
        dfg = get_dfg_graph(log, activity_key=activity_key,
                            timestamp_key=timestamp_key,
                            case_id_glue=case_id_key)
        from pm4py.statistics.start_activities.pandas import get as start_activities_module
        from pm4py.statistics.end_activities.pandas import get as end_activities_module
        start_activities = start_activities_module.get_start_activities(log, parameters=properties)
        end_activities = end_activities_module.get_end_activities(log, parameters=properties)
    else:
        from pm4py.algo.discovery.dfg import algorithm as dfg_discovery
        dfg = dfg_discovery.apply(log, parameters=get_properties(log))
        from pm4py.statistics.start_activities.log import get as start_activities_module
        from pm4py.statistics.end_activities.log import get as end_activities_module
        start_activities = start_activities_module.get_start_activities(log, parameters=get_properties(log))
        end_activities = end_activities_module.get_end_activities(log, parameters=get_properties(log))
    return dfg, start_activities, end_activities


def discover_directly_follows_graph(log: Union[EventLog, pd.DataFrame]) -> Tuple[dict, dict, dict]:
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    return discover_dfg(log)


def discover_performance_dfg(log: Union[EventLog, pd.DataFrame], business_hours: bool = False, worktiming: List[int] = [7, 17], weekends: List[int] = [6, 7], workcalendar=constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR) -> Tuple[dict, dict, dict]:
    """
    Discovers a performance directly-follows graph from an event log

    Parameters
    ---------------
    log
        Event log
    business_hours
        Enables/disables the computation based on the business hours (default: False)
    worktiming
        (If the business hours are enabled) The hour range in which the resources of the log are working (default: 7 to 17)
    weekends
        (If the business hours are enabled) The weekends days (default: Saturday (6), Sunday (7))

    Returns
    ---------------
    performance_dfg
        Performance DFG
    start_activities
        Start activities
    end_activities
        End activities
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)
        from pm4py.util import constants
        properties = get_properties(log)
        from pm4py.algo.discovery.dfg.adapters.pandas.df_statistics import get_dfg_graph
        activity_key = properties[constants.PARAMETER_CONSTANT_ACTIVITY_KEY] if constants.PARAMETER_CONSTANT_ACTIVITY_KEY in properties else xes_constants.DEFAULT_NAME_KEY
        timestamp_key = properties[constants.PARAMETER_CONSTANT_TIMESTAMP_KEY] if constants.PARAMETER_CONSTANT_TIMESTAMP_KEY in properties else xes_constants.DEFAULT_TIMESTAMP_KEY
        case_id_key = properties[constants.PARAMETER_CONSTANT_CASEID_KEY] if constants.PARAMETER_CONSTANT_CASEID_KEY in properties else constants.CASE_CONCEPT_NAME
        dfg = get_dfg_graph(log, activity_key=activity_key, timestamp_key=timestamp_key, case_id_glue=case_id_key, measure="performance", perf_aggregation_key="all",
                            business_hours=business_hours, worktiming=worktiming, weekends=weekends, workcalendar=workcalendar)
        from pm4py.statistics.start_activities.pandas import get as start_activities_module
        from pm4py.statistics.end_activities.pandas import get as end_activities_module
        start_activities = start_activities_module.get_start_activities(log, parameters=properties)
        end_activities = end_activities_module.get_end_activities(log, parameters=properties)
    else:
        from pm4py.algo.discovery.dfg.variants import performance as dfg_discovery
        properties = get_properties(log)
        properties[dfg_discovery.Parameters.AGGREGATION_MEASURE] = "all"
        properties[dfg_discovery.Parameters.BUSINESS_HOURS] = business_hours
        properties[dfg_discovery.Parameters.WORKTIMING] = worktiming
        properties[dfg_discovery.Parameters.WEEKENDS] = weekends
        dfg = dfg_discovery.apply(log, parameters=properties)
        from pm4py.statistics.start_activities.log import get as start_activities_module
        from pm4py.statistics.end_activities.log import get as end_activities_module
        start_activities = start_activities_module.get_start_activities(log, parameters=properties)
        end_activities = end_activities_module.get_end_activities(log, parameters=properties)
    return dfg, start_activities, end_activities


def discover_petri_net_alpha(log: Union[EventLog, pd.DataFrame]) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using the Alpha Miner

    Parameters
    --------------
    log
        Event log

    Returns
    --------------
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.alpha import algorithm as alpha_miner
    return alpha_miner.apply(log, variant=alpha_miner.Variants.ALPHA_VERSION_CLASSIC, parameters=get_properties(log))


def discover_petri_net_alpha_plus(log: Union[EventLog, pd.DataFrame]) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using the Alpha+ algorithm

    Parameters
    --------------
    log
        Event log

    Returns
    --------------
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.alpha import algorithm as alpha_miner
    return alpha_miner.apply(log, variant=alpha_miner.Variants.ALPHA_VERSION_PLUS, parameters=get_properties(log))


def discover_petri_net_inductive(log: Union[EventLog, pd.DataFrame], noise_threshold: float = 0.0) -> Tuple[
    PetriNet, Marking, Marking]:
    """
    Discovers a Petri net using the IMDFc algorithm

    Parameters
    --------------
    log
        Event log
    noise_threshold
        Noise threshold (default: 0.0)

    Returns
    --------------
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    pt = discover_process_tree_inductive(log, noise_threshold)
    from pm4py.convert import convert_to_petri_net
    return convert_to_petri_net(pt)


def discover_petri_net_heuristics(log: Union[EventLog, pd.DataFrame], dependency_threshold: float = 0.5,
                                  and_threshold: float = 0.65,
                                  loop_two_threshold: float = 0.5) -> Tuple[PetriNet, Marking, Marking]:
    """
    Discover a Petri net using the Heuristics Miner

    Parameters
    ---------------
    log
        Event log
    dependency_threshold
        Dependency threshold (default: 0.5)
    and_threshold
        AND threshold (default: 0.65)
    loop_two_threshold
        Loop two threshold (default: 0.5)

    Returns
    --------------
    petri_net
        Petri net
    initial_marking
        Initial marking
    final_marking
        Final marking
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.heuristics import algorithm as heuristics_miner
    heu_parameters = heuristics_miner.Variants.CLASSIC.value.Parameters
    parameters = get_properties(log)
    parameters[heu_parameters.DEPENDENCY_THRESH] = dependency_threshold
    parameters[heu_parameters.AND_MEASURE_THRESH] = and_threshold
    parameters[heu_parameters.LOOP_LENGTH_TWO_THRESH] = loop_two_threshold

    return heuristics_miner.apply(log, variant=heuristics_miner.Variants.CLASSIC, parameters=parameters)


def discover_process_tree_inductive(log: Union[EventLog, pd.DataFrame], noise_threshold: float = 0.0) -> ProcessTree:
    """
    Discovers a process tree using the IM algorithm

    Parameters
    --------------
    log
        Event log
    noise_threshold
        Noise threshold (default: 0.0)

    Returns
    --------------
    process_tree
        Process tree object
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.inductive import algorithm as inductive_miner
    parameters = get_properties(log)
    parameters[inductive_miner.Variants.IM_CLEAN.value.Parameters.NOISE_THRESHOLD] = noise_threshold
    return inductive_miner.apply_tree(log, variant=inductive_miner.Variants.IM_CLEAN, parameters=parameters)


@deprecation.deprecated(deprecated_in='2.2.2', removed_in='2.4.0',
                        details='discover_tree_inductive is deprecated, use discover_process_tree_inductive')
def discover_tree_inductive(log: Union[EventLog, pd.DataFrame], noise_threshold: float = 0.0) -> ProcessTree:
    warnings.warn('discover_tree_inductive is deprecated, use discover_process_tree_inductive', DeprecationWarning)
    """
    Discovers a process tree using the IMDFc algorithm

    Parameters
    --------------
    log
        Event log
    noise_threshold
        Noise threshold (default: 0.0)

    Returns
    --------------
    process_tree
        Process tree object
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    return discover_process_tree_inductive(log, noise_threshold)


def discover_heuristics_net(log: Union[EventLog, pd.DataFrame], dependency_threshold: float = 0.5,
                            and_threshold: float = 0.65,
                            loop_two_threshold: float = 0.5) -> HeuristicsNet:
    """
    Discovers an heuristics net

    Parameters
    ---------------
    log
        Event log
    dependency_threshold
        Dependency threshold (default: 0.5)
    and_threshold
        AND threshold (default: 0.65)
    loop_two_threshold
        Loop two threshold (default: 0.5)

    Returns
    --------------
    heu_net
        Heuristics net
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.heuristics import algorithm as heuristics_miner
    heu_parameters = heuristics_miner.Variants.CLASSIC.value.Parameters
    parameters = get_properties(log)
    parameters[heu_parameters.DEPENDENCY_THRESH] = dependency_threshold
    parameters[heu_parameters.AND_MEASURE_THRESH] = and_threshold
    parameters[heu_parameters.LOOP_LENGTH_TWO_THRESH] = loop_two_threshold
    return heuristics_miner.apply_heu(log, variant=heuristics_miner.Variants.CLASSIC, parameters=parameters)


def derive_minimum_self_distance(log: Union[DataFrame, EventLog, EventStream]) -> Dict[str, int]:
    '''
        This algorithm computes the minimum self-distance for each activity observed in an event log.
        The self distance of a in <a> is infinity, of a in <a,a> is 0, in <a,b,a> is 1, etc.
        The activity key 'concept:name' is used.


        Parameters
        ----------
        log
            event log (either pandas.DataFrame, EventLog or EventStream)

        Returns
        -------
            dict mapping an activity to its self-distance, if it exists, otherwise it is not part of the dict.
        '''
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    from pm4py.algo.discovery.minimum_self_distance import algorithm as msd
    return msd.apply(log, parameters=get_properties(log))


def discover_footprints(*args: Union[EventLog, Tuple[PetriNet, Marking, Marking], ProcessTree]) -> Union[
    List[Dict[str, Any]], Dict[str, Any]]:
    """
    Discovers the footprints out of the provided event log / pocess model

    Parameters
    --------------
    args
        Event log / process model
    """
    from pm4py.algo.discovery.footprints import algorithm as fp_discovery
    return fp_discovery.apply(*args)


def discover_eventually_follows_graph(log: Union[EventLog, pd.DataFrame]) -> Dict[Tuple[str, str], int]:
    """
    Gets the eventually follows graph from a log object

    Parameters
    ---------------
    log
        Log object

    Returns
    ---------------
    eventually_follows_graph
        Dictionary of tuples of activities that eventually follows each other; along with the number of occurrences
    """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    if check_is_pandas_dataframe(log):
        check_pandas_dataframe_columns(log)
        from pm4py.statistics.eventually_follows.pandas import get
        return get.apply(log, parameters=get_properties(log))
    else:
        from pm4py.statistics.eventually_follows.log import get
        return get.apply(log, parameters=get_properties(log))


def discover_bpmn_inductive(log: Union[EventLog, pd.DataFrame], noise_threshold: float = 0.0) -> BPMN:
    """
        Discovers a BPMN using the Inductive Miner algorithm

        Parameters
        --------------
        log
            Event log
        noise_threshold
            Noise threshold (default: 0.0)

        Returns
        --------------
        bpmn_diagram
            BPMN diagram
        """
    if type(log) not in [pd.DataFrame, EventLog, EventStream]: raise Exception("the method can be applied only to a traditional event log!")

    pt = discover_process_tree_inductive(log, noise_threshold)
    from pm4py.convert import convert_to_bpmn
    return convert_to_bpmn(pt)


def discover_ocdfg(ocel: OCEL, business_hours=False, worktiming=[7, 17], weekends=[6, 7]) -> Dict[str, Any]:
    """
    Discovers an OC-DFG from an object-centric event log.

    Reference paper:
    Berti, Alessandro, and Wil van der Aalst. "Extracting multiple viewpoint models from relational databases." Data-Driven Process Discovery and Analysis. Springer, Cham, 2018. 24-51.


    Parameters
    ----------------
    ocel
        Object-centric event log
    business_hours
        Boolean value that enables the usage of the business hours
    worktiming
        (if business hours are in use) work timing during the day (default: [7, 17])
    weekends
        (if business hours are in use) weekends (default: [6, 7])

    Returns
    ---------------
    ocdfg
        Object-centric directly-follows graph
    """
    parameters = {}
    parameters["business_hours"] = business_hours
    parameters["worktiming"] = worktiming
    parameters["weekends"] = weekends
    from pm4py.algo.discovery.ocel.ocdfg import algorithm as ocdfg_discovery
    return ocdfg_discovery.apply(ocel, parameters=parameters)


def discover_oc_petri_net(ocel: OCEL) -> Dict[str, Any]:
    """
    Discovers an object-centric Petri net from the provided object-centric event log.

    Reference paper: van der Aalst, Wil MP, and Alessandro Berti. "Discovering object-centric Petri nets." Fundamenta informaticae 175.1-4 (2020): 1-40.

    Parameters
    -----------------
    ocel
        Object-centric event log

    Returns
    ----------------
    ocpn
        Object-centric Petri net
    """
    from pm4py.algo.discovery.ocel.ocpn import algorithm as ocpn_discovery
    return ocpn_discovery.apply(ocel)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum

from pm4py.objects.petri_net.importer.variants import pnml
from pm4py.util import exec_utils


class Variants(Enum):
    PNML = pnml


PNML = Variants.PNML


def apply(input_file_path, variant=PNML, parameters=None):
    """
    Import a Petri net from a PNML file

    Parameters
    ------------
    input_file_path
        Input file path
    parameters
        Other parameters of the importer
    variant
        Variant of the algorithm to use, possible values:
            - Variants.PNML
    """
    return exec_utils.get_variant(variant).import_net(input_file_path, parameters=parameters)


def deserialize(petri_string, variant=PNML, parameters=None):
    """
    Deserialize a text/binary string representing a Petri net in the PNML format

    Parameters
    ----------
    petri_string
        Petri net expressed as PNML string
    variant
        Variant of the algorithm to use, possible values:
            - Variants.PNML
    parameters
        Other parameters of the algorithm
    """
    return exec_utils.get_variant(variant).import_net_from_string(petri_string, parameters=parameters)



import pm4py
from pm4py.algo.decision_mining import algorithm as decision_mining
import os


def execute_script():
    log_path = os.path.join("..", "tests", "input_data", "roadtraffic100traces.xes")
    log = pm4py.read_xes(log_path)
    net, im, fm = pm4py.discover_petri_net_inductive(log)
    net, im, fm = decision_mining.create_data_petri_nets_with_decisions(log, net, im, fm)
    pm4py.view_petri_net(net, im, fm, format="svg")


if __name__ == "__main__":
    execute_script()



from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.objects.log.exporter.xes import exporter as xes_exporter
from tests.constants import INPUT_DATA_DIR, OUTPUT_DATA_DIR, PROBLEMATIC_XES_DIR, COMPRESSED_INPUT_DATA
import logging
import unittest
import os


class XesImportExportTest(unittest.TestCase):
    def test_importExportXEStoXES(self):
        # to avoid static method warnings in tests,
        # that by construction of the unittest package have to be expressed in such way
        self.dummy_variable = "dummy_value"
        log = xes_importer.apply(os.path.join(INPUT_DATA_DIR, "running-example.xes"))
        xes_exporter.apply(log, os.path.join(OUTPUT_DATA_DIR, "running-example-exported.xes"))
        log_imported_after_export = xes_importer.apply(
            os.path.join(OUTPUT_DATA_DIR, "running-example-exported.xes"))
        self.assertEqual(len(log), len(log_imported_after_export))
        os.remove(os.path.join(OUTPUT_DATA_DIR, "running-example-exported.xes"))

    def test_importExportProblematicLogs(self):
        # to avoid static method warnings in tests,
        # that by construction of the unittest package have to be expressed in such way
        self.dummy_variable = "dummy_value"
        logs = os.listdir(PROBLEMATIC_XES_DIR)
        for log in logs:
            log_full_path = os.path.join(PROBLEMATIC_XES_DIR, log)
            try:
                output_log_path = os.path.join(OUTPUT_DATA_DIR, log)
                log = xes_importer.apply(log_full_path)
                xes_exporter.apply(log, output_log_path)
                log_imported_after_export = xes_importer.apply(output_log_path)
                self.assertEqual(len(log), len(log_imported_after_export))
                os.remove(output_log_path)
            except SyntaxError as e:
                logging.info("SyntaxError on log " + str(log) + ": " + str(e))

    def test_importExportXESfromGZIP_imp1(self):
        # to avoid static method warnings in tests,
        # that by construction of the unittest package have to be expressed in such way
        self.dummy_variable = "dummy_value"
        log = xes_importer.apply(os.path.join(COMPRESSED_INPUT_DATA, "01_running-example.xes.gz"))
        xes_exporter.apply(log, os.path.join(OUTPUT_DATA_DIR, "01-running-example.xes"),
                           parameters={xes_exporter.Variants.ETREE.value.Parameters.COMPRESS: True})
        os.remove(os.path.join(OUTPUT_DATA_DIR, "01-running-example.xes.gz"))

    def test_importXESfromGZIP_imp2(self):
        # to avoid static method warnings in tests,
        # that by construction of the unittest package have to be expressed in such way
        self.dummy_variable = "dummy_value"
        log = xes_importer.apply(os.path.join(COMPRESSED_INPUT_DATA, "01_running-example.xes.gz"))
        del log


if __name__ == "__main__":
    unittest.main()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import tempfile
from copy import copy

from graphviz import Digraph

from pm4py.statistics.attributes.log import get as attr_get
from pm4py.objects.dfg.utils import dfg_utils
from pm4py.util import xes_constants as xes
from pm4py.visualization.common.utils import *
from pm4py.util import exec_utils
from pm4py.statistics.sojourn_time.log import get as soj_time_get
from pm4py.util import constants
from enum import Enum

from typing import Optional, Dict, Any, Tuple
from pm4py.objects.log.obj import EventLog


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    FORMAT = "format"
    MAX_NO_EDGES_IN_DIAGRAM = "maxNoOfEdgesInDiagram"
    START_ACTIVITIES = "start_activities"
    END_ACTIVITIES = "end_activities"
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    FONT_SIZE = "font_size"
    AGGREGATION_MEASURE = "aggregation_measure"
    BGCOLOR = "bgcolor"
    STAT_LOCALE = "stat_locale"

def get_min_max_value(dfg):
    """
    Gets min and max value assigned to edges
    in DFG graph

    Parameters
    -----------
    dfg
        Directly follows graph

    Returns
    -----------
    min_value
        Minimum value in directly follows graph
    max_value
        Maximum value in directly follows graph
    """
    min_value = 9999999999
    max_value = -1

    for edge in dfg:
        if dfg[edge] < min_value:
            min_value = dfg[edge]
        if dfg[edge] > max_value:
            max_value = dfg[edge]

    return min_value, max_value


def assign_penwidth_edges(dfg):
    """
    Assign penwidth to edges in directly-follows graph

    Parameters
    -----------
    dfg
        Direcly follows graph

    Returns
    -----------
    penwidth
        Graph penwidth that edges should have in the direcly follows graph
    """
    penwidth = {}
    min_value, max_value = get_min_max_value(dfg)
    for edge in dfg:
        v0 = dfg[edge]
        v1 = get_arc_penwidth(v0, min_value, max_value)
        penwidth[edge] = str(v1)

    return penwidth


def get_activities_color_soj_time(soj_time):
    """
    Gets the color for the activities based on the sojourn time

    Parameters
    ----------------
    soj_time
        Sojourn time

    Returns
    ----------------
    act_color
        Dictionary associating each activity to a color based on the sojourn time
    """
    activities_color = {}

    min_soj_time, max_soj_time = get_min_max_value(soj_time)

    for ac in soj_time:
        act_soj_time = soj_time[ac]

        trans_base_color = int(
            255 - 100 * (act_soj_time - min_soj_time) / (max_soj_time - min_soj_time + 0.00001))
        trans_base_color_hex = str(hex(trans_base_color))[2:].upper()

        activities_color[ac] = "#" + "FF" + trans_base_color_hex + trans_base_color_hex

    return activities_color

def get_edges_color(duration_list: list) -> str:
    """
    Gets the color for the activities based on the sojourn time

    Parameters
    ----------------
    soj_time
        Sojourn time

    Returns
    ----------------
    act_color
        Dictionary associating each activity to a color based on the sojourn time
    """
    LIGHTEST_COLOR = 55
    edges_color = {}

    min_time, max_time = get_min_max_value(duration_list)
    min_color = 255 - LIGHTEST_COLOR

    for ac in duration_list:
        current_time = duration_list[ac]

        trans_base_color = int(min_color - 
            min_color * (current_time - min_time) / 
                (max_time - min_time + 0.00001))
        trans_base_color_hex = str(hex(trans_base_color))[2:].upper()

        if len(trans_base_color_hex) == 1:
            trans_base_color_hex = "0" + trans_base_color_hex

        edges_color[ac] = "#" + 3 * trans_base_color_hex

    return edges_color

def graphviz_visualization(activities_count, dfg, image_format="png", measure="frequency",
                           max_no_of_edges_in_diagram=100000, start_activities=None, 
                           end_activities=None, soj_time=None, font_size="12", 
                           bgcolor="transparent", stat_locale: dict = None):
    """
    Do GraphViz visualization of a DFG graph

    Parameters
    -----------
    activities_count
        Count of attributes in the log (may include attributes that are not in the DFG graph)
    dfg
        DFG graph
    image_format
        GraphViz should be represented in this format
    measure
        Describes which measure is assigned to edges in directly follows graph (frequency/performance)
    max_no_of_edges_in_diagram
        Maximum number of edges in the diagram allowed for visualization
    start_activities
        Start activities of the log
    end_activities
        End activities of the log
    soj_time
        For each activity, the sojourn time in the log
    stat_locale
        Dict to locale the stat strings
    
    Returns
    -----------
    viz
        Digraph object
    """
    if start_activities is None:
        start_activities = []
    if end_activities is None:
        end_activities = []
    if stat_locale is None:
        stat_locale = {}

    filename = tempfile.NamedTemporaryFile(suffix='.gv')
    viz = Digraph("", filename=filename.name, engine='dot', graph_attr={'bgcolor': bgcolor})

    # first, remove edges in diagram that exceeds the maximum number of edges in the diagram
    dfg_key_value_list = []
    for edge in dfg:
        dfg_key_value_list.append([edge, dfg[edge]])
    # more fine grained sorting to avoid that edges that are below the threshold are
    # undeterministically removed
    dfg_key_value_list = sorted(dfg_key_value_list, key=lambda x: (x[1], x[0][0], x[0][1]), reverse=True)
    dfg_key_value_list = dfg_key_value_list[0:min(len(dfg_key_value_list), max_no_of_edges_in_diagram)]
    dfg_allowed_keys = [x[0] for x in dfg_key_value_list]
    dfg_keys = list(dfg.keys())
    for edge in dfg_keys:
        if edge not in dfg_allowed_keys:
            del dfg[edge]

    # calculate edges penwidth
    penwidth = assign_penwidth_edges(dfg)
    activities_in_dfg = set()
    activities_count_int = copy(activities_count)

    for edge in dfg:
        activities_in_dfg.add(edge[0])
        activities_in_dfg.add(edge[1])

    # assign attributes color
    activities_color = get_activities_color_soj_time(soj_time)
    edges_color = get_edges_color(dfg)

    # represent nodes
    viz.attr('node', shape='box')

    if len(activities_in_dfg) == 0:
        activities_to_include = sorted(list(set(activities_count_int)))
    else:
        # take unique elements as a list not as a set (in this way, nodes are added in the same order to the graph)
        activities_to_include = sorted(list(set(activities_in_dfg)))

    activities_map = {}

    for act in activities_to_include:
        if "frequency" in measure and act in activities_count_int:
            viz.node(str(hash(act)), act + " (" + str(activities_count_int[act]) + ")", style='filled',
                     fillcolor=activities_color[act], fontsize=font_size)
            activities_map[act] = str(hash(act))
        else:
            node_label = act
            if soj_time[act] > -1:
                stat_string = human_readable_stat(soj_time[act])
                node_label = f"{act} ({stat_string})"
            viz.node(str(hash(act)), node_label, fontsize=font_size,
                     style='filled', fillcolor=activities_color[act])
            activities_map[act] = str(hash(act))

    # make edges addition always in the same order
    dfg_edges = sorted(list(dfg.keys()))

    # represent edges
    for edge in dfg_edges:
        if "frequency" in measure:
            label = str(dfg[edge])
        else:
            label = human_readable_stat(dfg[edge], stat_locale)
        viz.edge(str(hash(edge[0])), str(hash(edge[1])), label=label, 
                 color=edges_color[edge], style = "bold",
                 penwidth=str(penwidth[edge]), fontsize=font_size)

    start_activities_to_include = [act for act in start_activities if act in activities_map]
    end_activities_to_include = [act for act in end_activities if act in activities_map]

    if start_activities_to_include:
        viz.node("@@startnode", "<&#9679;>", shape='circle', fontsize="34")
        for act in start_activities_to_include:
            label = str(start_activities[act]) if isinstance(start_activities, dict) else ""
            viz.edge("@@startnode", activities_map[act], label=label, fontsize=font_size)

    if end_activities_to_include:
        # <&#9632;>
        viz.node("@@endnode", "<&#9632;>", shape='doublecircle', fontsize="32")
        for act in end_activities_to_include:
            label = str(end_activities[act]) if isinstance(end_activities, dict) else ""
            viz.edge(activities_map[act], "@@endnode", label=label, fontsize=font_size)

    viz.attr(overlap='false')
    viz.attr(fontsize='11')

    viz.format = image_format

    return viz


def apply(dfg: Dict[Tuple[str, str], int], log: EventLog = None, parameters: Optional[Dict[Any, Any]] = None, activities_count : Dict[str, int] = None, soj_time: Dict[str, float] = None) -> Digraph:
    """
    Visualize a performance directly-follows graph

    Parameters
    -----------------
    dfg
        Performance Directly-follows graph
    log
        (if provided) Event log for the calculation of statistics
    activities_count
        (if provided) Dictionary associating to each activity the number of occurrences in the log.
    soj_time
        (if provided) Dictionary associating to each activity the average sojourn time
    parameters
        Variant-specific parameters

    Returns
    -----------------
    gviz
        Graphviz digraph
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes.DEFAULT_NAME_KEY)
    image_format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    max_no_of_edges_in_diagram = exec_utils.get_param_value(Parameters.MAX_NO_EDGES_IN_DIAGRAM, parameters, 100000)
    start_activities = exec_utils.get_param_value(Parameters.START_ACTIVITIES, parameters, [])
    end_activities = exec_utils.get_param_value(Parameters.END_ACTIVITIES, parameters, [])
    font_size = exec_utils.get_param_value(Parameters.FONT_SIZE, parameters, 12)
    font_size = str(font_size)
    activities = dfg_utils.get_activities_from_dfg(dfg)
    aggregation_measure = exec_utils.get_param_value(Parameters.AGGREGATION_MEASURE, parameters, "mean")
    bgcolor = exec_utils.get_param_value(Parameters.BGCOLOR, parameters, "transparent")
    stat_locale = exec_utils.get_param_value(Parameters.STAT_LOCALE, parameters, None)

    if stat_locale is None:
        stat_locale = {}

    if activities_count is None:
        if log is not None:
            activities_count = attr_get.get_attribute_values(log, activity_key, parameters=parameters)
        else:
            # the frequency of an activity in the log is at least the number of occurrences of
            # incoming arcs in the DFG.
            # if the frequency of the start activities nodes is also provided, use also that.
            activities_count = Counter({key: 0 for key in activities})
            for el in dfg:
                activities_count[el[1]] += dfg[el]
            if isinstance(start_activities, dict):
                for act in start_activities:
                    activities_count[act] += start_activities[act]

    if soj_time is None:
        if log is not None:
            soj_time = soj_time_get.apply(log, parameters=parameters)
        else:
            soj_time = {key: -1 for key in activities}

    # if all the aggregation measures are provided for a given key,
    # then pick one of the values for the representation
    dfg0 = dfg
    dfg = {}
    for key in dfg0:
        try:
            if aggregation_measure in dfg0[key]:
                dfg[key] = dfg0[key][aggregation_measure]
            else:
                dfg[key] = dfg0[key]
        except:
            dfg[key] = dfg0[key]

    return graphviz_visualization(activities_count, dfg, image_format=image_format, measure="performance",
                                  max_no_of_edges_in_diagram=max_no_of_edges_in_diagram,
                                  start_activities=start_activities, end_activities=end_activities, 
                                  soj_time=soj_time, font_size=font_size, bgcolor=bgcolor, 
                                  stat_locale=stat_locale)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.log import obj
from copy import copy, deepcopy
import logging


def show_nice_log(old_log):
    nl = []
    for trace in old_log:
        nt = []
        for element in trace:
            nt.append(element['concept:name'])
        nl.append(nt)
    return nl


def empty_trace(l):
    # checks if there are empty traces in the log, if so, creates new_log without those empty traces
    contains_empty_trace = False
    for trace in l:
        if len(trace) == 0:
            contains_empty_trace = True

    if contains_empty_trace:
        new_log = obj.EventLog()
        for trace in l:
            if len(trace) != 0:
                new_log.append(trace)
        return True, new_log
    else:
        return False, l


def filter_activity_from_log(l, act, activity_key):
    # remove the activity from every trace in the log
    # as trace doesnt have remove function, we just create new traces without chosen_activity
    act_str = str(act)
    new_log = obj.EventLog()
    for trace in l:
        new_trace = obj.Trace()
        for event in trace:
            if not event[activity_key] == act_str:
                new_trace.append(event)
        new_log.append(new_trace)

    return new_log


def index_containing(l, activities, activity_key):
    ret = {act: [] for act in activities}
    for index, trace in enumerate(l):
        activities = set(ev[activity_key] for ev in trace)
        for act in activities:
            ret[act].append(index)

    return ret


def filter_activity_use_idx(l, act, activity_key, idx):
    act_str = str(act)
    i_act = idx[act]
    new_log = obj.EventLog()
    i = 0
    j = 0
    while i < len(l):
        if j < len(i_act) and i == i_act[j]:
            new_trace = obj.Trace()
            for event in l[i]:
                if not event[activity_key] == act_str:
                    new_trace.append(event)
            new_log.append(new_trace)
            j = j + 1
        else:
            new_log.append(l[i])
        i = i + 1
    return new_log


def act_once_per_trace(l, activities, activity_key):
    small_log = obj.EventLog()
    small_trace = obj.Trace()
    new_log = obj.EventLog()
    number_of_traces = len(l)
    possible_activities = list()
    # transform dict of activities to list
    activities_dict = activities
    for key, value in activities_dict.items():
        # if activity appears as often as there are traces, add to list of possible activities:
        if value == number_of_traces:
            possible_activities.append(key)

    chosen_activity = None
    # find an activity that appears exactly once per trace and save it in chose_activity
    for act in possible_activities:
        fits_log = True
        for trace in l:
            fits_trace = False
            for element in trace:
                # enough to check if element occurs once per trace as number of occurrences equals the number of traces
                if act == element[activity_key]:
                    fits_trace = True
            if not fits_trace:
                fits_log = False

        if fits_log:
            chosen_activity = act
            break

    # save the chosen activity in a new trace, so that it can later be appended as leaf to our subtree
    for trace in l:
        if len(small_trace) > 0:
            break
        for element in trace:
            if element[activity_key] == chosen_activity:
                small_trace.append(element)
                small_log.append(small_trace)
                break

    if chosen_activity is not None:
        new_log = filter_activity_from_log(l, chosen_activity, activity_key)
        logging_output = "activity once per trace: " + str(chosen_activity)
        logging.debug(logging_output)
        return True, new_log, small_log
    else:
        return False, new_log, chosen_activity


def activity_concurrent(self, l, activities, activity_key, parameters=None):
    from pm4py.algo.discovery.inductive.variants.im.data_structures import subtree_plain as subtree

    small_log = obj.EventLog()
    test_log = obj.EventLog()
    key = None
    activities_copy = copy(activities)
    empty_trace = obj.Trace()
    idx = index_containing(l, activities, activity_key)

    for key, value in activities_copy.items():  # iterate through activities (saved in key)

        test_log = filter_activity_use_idx(l, key, activity_key, idx)
        #test_log = filter_activity_from_log(l, key, activity_key)
        # unsure about this one:
        contains_empty_trace = False
        for trace in test_log:
            if len(trace) == 0:
                contains_empty_trace = True
        if contains_empty_trace:
            continue

        # more efficient deepcopy
        self_copy = deepcopy(self)
        cut = subtree.SubtreePlain.check_for_cut(self_copy, test_log,
                                                 key, parameters=parameters)  # check if leaving out act, leads to finding cut
        if cut:
            # save act to small_trace, so that it can be appended as leaf later on
            for trace in l:
                small_trace = obj.Trace()
                contains_activity = False
                for element in trace:
                    if element[activity_key] == key:
                        contains_activity = True
                        small_trace.append(element)
                small_log.append(small_trace)

                if not contains_activity:
                    small_log.append(empty_trace)
            logging_output = "activity concurrent: " + str(key)
            logging.debug(logging_output)
            return True, test_log, small_log, key  # if so, return new log

    return False, test_log, small_log, key  # if,  after iterating through all act's still no cut is found, return false


def split_between_end_and_start(trace, start_activities, end_activities, activity_key):
    # splits a trace between the first occurrence of an end activity  following a start activity
    found_split = False
    new_trace_1 = obj.Trace()
    new_trace_2 = obj.Trace()
    i = 0

    while not found_split and i < len(trace) - 1:
        if trace[i][activity_key] in end_activities and trace[i + 1][activity_key] in start_activities:
            found_split = True
            j = 0
            while j <= i:
                new_trace_1.append(trace[j])
                j += 1
            for k in range(i + 1, len(trace)):
                new_trace_2.append(trace[k])
            break
        else:
            i += 1
    if not found_split:
        new_trace_1 = trace
    return new_trace_1, new_trace_2, found_split


def strict_tau_loop(l, start_activities, end_activities, activity_key):
    new_log = obj.EventLog()
    for trace in l:  # for all traces
        t1, t2, found_split = split_between_end_and_start(trace, start_activities, end_activities,
                                                          activity_key)  # look for split
        if found_split:
            new_log.append(t1)
            while found_split:  # if split is found
                t1, t2, found_split = split_between_end_and_start(t2, start_activities,
                                                                  end_activities, activity_key)  # continue to split
                new_log.append(t1)
        else:
            new_log.append(trace)  # if there is nothing to split, append the whole trace

    if len(new_log) > len(l):
        logging.debug("strict_tau_loop")
        return True, new_log
    else:
        return False, new_log


def split_before_start(trace, start_activities, activity_key):
    # if there is only one activity, there is nothing to split
    if len(trace) == 1:
        return trace, trace, False
    # if none of the above cases apply, we split at the occurence of a start activity
    found_split = False
    new_trace_1 = obj.Trace()
    new_trace_2 = obj.Trace()
    i = 1
    while not found_split and i < len(trace):  # for all events in trace
        if trace[i][activity_key] in start_activities and len(trace) > 1:
            found_split = True
            for j in range(0, i):
                new_trace_1.append(trace[j])
            for k in range(i, len(trace)):
                new_trace_2.append(trace[k])
        i += 1

    return new_trace_1, new_trace_2, found_split


def tau_loop(l, start_activities, activity_key):
    # pretty much the same code as in strict_tau_loop, just that we split at a different point
    new_log = obj.EventLog()
    for trace in l:
        t1, t2, found_split = split_before_start(trace, start_activities, activity_key)
        if found_split and len(t2) != 0:
            new_log.append(t1)
            while found_split:
                t2_backup = copy(t2)
                t1, t2, found_split = split_before_start(t2, start_activities, activity_key)
                if len(t1) != 0:
                    new_log.append(t1)
                else:
                    new_log.append(t2_backup)
        else:
            new_log.append(trace)

    if len(new_log) > len(l):
        logging.debug("tau_loop")
        return True, new_log
    else:
        return False, new_log



from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.algo.discovery.transition_system import algorithm as ts_discovery
from pm4py.visualization.transition_system import visualizer as ts_vis
import os


def execute_script():
    # log_path = "C:/Users/bas/Documents/tue/svn/private/logs/ilp_test_2_abcd_acbd.xes"
    log_path = os.path.join("..", "tests", "input_data", "running-example.xes")
    log = xes_importer.apply(log_path)
    ts = ts_discovery.apply(log, parameters={"include_data": True})
    viz = ts_vis.apply(ts, parameters={ts_vis.Variants.VIEW_BASED.value.Parameters.FORMAT: "svg"})
    ts_vis.view(viz)
    for state in ts.states:
        print(state.name, "ingoing=", len(state.data["ingoing_events"]), "outgoing=",
              len(state.data["outgoing_events"]))
    for trans in ts.transitions:
        print(trans.name, trans.from_state.name, trans.to_state.name, "frequency=", len(trans.data["events"]))


if __name__ == '__main__':
    execute_script()



import os
import unittest


class Pm4pyImportPackageTest(unittest.TestCase):
    def test_importeverything(self):
        # to avoid static method warnings in tests,
        # that by construction of the unittest package have to be expressed in such way
        self.dummy_variable = "dummy_value"
        import pm4py
        log = pm4py.objects.log.importer.xes.importer.apply(
            os.path.join("input_data", "running-example.xes"))
        self.assertEqual(len(log), 6)


if __name__ == "__main__":
    unittest.main()



from pm4py.objects.log.importer.xes import importer
from pm4py.algo.filtering.log.variants import variants_filter
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from pm4py.objects.conversion.process_tree import converter
from pm4py.algo.discovery.footprints import algorithm as footprints_discovery
from pm4py.algo.conformance.footprints import algorithm as footprints_conformance
from pm4py.visualization.footprints import visualizer as fp_visualizer
import os
from copy import deepcopy


def execute_script():
    # import a log
    log = importer.apply(os.path.join("..", "tests", "input_data", "receipt.xes"))
    # found a filtered version of the log that is used to discover a process model
    filtered_log = variants_filter.filter_log_variants_percentage(log, 0.2)
    # discover a process tree using inductive miner
    tree = inductive_miner.apply_tree(filtered_log)
    print(tree)
    # apply the conversion of a process tree into a Petri net
    net, im, fm = converter.apply(tree)
    # Footprints discovery: discover a list of footprints
    # for all the cases of the log
    fp_log = footprints_discovery.apply(log)
    # discover the footpritns from the process tree
    fp_tree = footprints_discovery.apply(tree)
    # discover the footpritns from the Petri net
    fp_net = footprints_discovery.apply(net, im)
    print(len(fp_tree["sequence"]), len(fp_tree["parallel"]), len(fp_net["sequence"]), len(fp_net["parallel"]))
    print(fp_tree["sequence"] == fp_net["sequence"] and fp_tree["parallel"] == fp_net["parallel"])
    # apply the footprints conformance checking
    conf = footprints_conformance.apply(fp_log, fp_net)
    for trace_an in conf:
        if trace_an:
            # print the first anomalous trace (containing deviations
            # that are contained in the trace but not allowed by the model)
            print(trace_an)
            break
    # finds the footprints for the entire log (not case-by-case, but taking
    # the relations that appear inside the entire log)
    fp_log_entire = footprints_discovery.apply(log, variant=footprints_discovery.Variants.ENTIRE_EVENT_LOG)
    # visualize the footprint table
    gviz = fp_visualizer.apply(fp_log_entire, fp_net, parameters={"format": "svg"})
    fp_visualizer.view(gviz)


if __name__ == "__main__":
    execute_script()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.process_tree import obj as pt


def transform_dfg_to_directed_nx_graph(dfg, alphabet):
    import networkx as nx

    nx_graph = nx.DiGraph()
    nx_graph.add_nodes_from(alphabet)
    for a, b in dfg:
        nx_graph.add_edge(a, b)
    return nx_graph


def __merge_groups_for_acts(a, b, groups):
    group_a = None
    group_b = None
    for group in groups:
        if a in group:
            group_a = group
        if b in group:
            group_b = group
    groups = [group for group in groups if group != group_a and group != group_b]
    groups.append(group_a.union(group_b))
    return groups


def __filter_dfg_on_threshold(dfg, end_activities, threshold):
    outgoing_max_occ = {}
    for x, y in dfg.items():
        act = x[0]
        if act not in outgoing_max_occ:
            outgoing_max_occ[act] = y
        else:
            outgoing_max_occ[act] = max(y, outgoing_max_occ[act])
        if act in end_activities:
            outgoing_max_occ[act] = max(outgoing_max_occ[act], end_activities[act])
    dfg_list = sorted([(x, y) for x, y in dfg.items()], key=lambda x: (x[1], x[0]), reverse=True)
    dfg_list = [x for x in dfg_list if x[1] > threshold * outgoing_max_occ[x[0][0]]]
    dfg_list = [x[0] for x in dfg_list]
    # filter the elements in the DFG
    dfg = {x: y for x, y in dfg.items() if x in dfg_list}
    return dfg


def __flower(alphabet, root):
    operator = pt.ProcessTree(operator=pt.Operator.LOOP, parent=root)
    operator.children.append(pt.ProcessTree(parent=operator))
    xor = pt.ProcessTree(operator=pt.Operator.XOR)
    operator.children.append(xor)
    for a in alphabet:
        tree = pt.ProcessTree(label=a, parent=xor)
        xor.children.append(tree)
    return operator


class DfgSaEaActCount(object):
    def __init__(self, dfg, sa, ea, act_count):
        self.dfg = dfg
        self.start_activities = sa
        self.end_activities = ea
        self.act_count = act_count

    def __str__(self):
        return str((self.dfg, self.start_activities, self.end_activities, self.act_count))

    def __repr__(self):
        return str((self.dfg, self.start_activities, self.end_activities, self.act_count))



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import uuid

from pm4py.objects.bpmn.obj import BPMN
from pm4py.util import constants


def apply(bpmn_graph, target_path, parameters=None):
    """
    Exports the BPMN diagram to a file

    Parameters
    -------------
    bpmn_graph
        BPMN diagram
    target_path
        Target path
    parameters
        Possible parameters of the algorithm
    """
    xml_string = get_xml_string(bpmn_graph, parameters=parameters)
    F = open(target_path, "wb")
    F.write(xml_string)
    F.close()


def get_xml_string(bpmn_graph, parameters=None):
    if parameters is None:
        parameters = {}

    import xml.etree.ElementTree as ET
    from xml.dom import minidom

    definitions = ET.Element("bpmn:definitions")
    definitions.set("xmlns:bpmn", "http://www.omg.org/spec/BPMN/20100524/MODEL")
    definitions.set("xmlns:bpmndi", "http://www.omg.org/spec/BPMN/20100524/DI")
    definitions.set("xmlns:omgdc", "http://www.omg.org/spec/DD/20100524/DC")
    definitions.set("xmlns:omgdi", "http://www.omg.org/spec/DD/20100524/DI")
    definitions.set("xmlns:xsi", "http://www.w3.org/2001/XMLSchema-instance")
    definitions.set("targetNamespace", "http://www.signavio.com/bpmn20")
    definitions.set("typeLanguage", "http://www.w3.org/2001/XMLSchema")
    definitions.set("expressionLanguage", "http://www.w3.org/1999/XPath")
    definitions.set("xmlns:xsd", "http://www.w3.org/2001/XMLSchema")

    diagram = ET.SubElement(definitions, "bpmndi:BPMNDiagram", {"id": "id" + str(uuid.uuid4()), "name": "diagram"})
    all_processes = set()
    process_planes = {}
    process_process = {}
    for node in bpmn_graph.get_nodes():
        all_processes.add(node.get_process())

    for process in all_processes:
        plane = ET.SubElement(diagram, "bpmndi:BPMNPlane",
                              {"bpmnElement": "id" + process, "id": "id" + str(uuid.uuid4())})
        process_planes[process] = plane

        p = ET.SubElement(definitions, "bpmn:process",
                          {"id": "id" + process, "isClosed": "false", "isExecutable": "false",
                           "processType": "None"})
        process_process[process] = p

    for node in bpmn_graph.get_nodes():
        process = node.get_process()

        node_shape = ET.SubElement(process_planes[process], "bpmndi:BPMNShape",
                                   {"bpmnElement": node.get_id(), "id": node.get_id() + "_gui"})
        node_shape_layout = ET.SubElement(node_shape, "omgdc:Bounds",
                                          {"height": str(node.get_height()), "width": str(node.get_width()),
                                           "x": str(node.get_x()),
                                           "y": str(node.get_y())})

    for flow in bpmn_graph.get_flows():
        process = flow.get_process()

        flow_shape = ET.SubElement(process_planes[process], "bpmndi:BPMNEdge",
                                   {"bpmnElement": "id" + str(flow.get_id()),
                                    "id": "id" + str(flow.get_id()) + "_gui"})
        for x, y in flow.get_waypoints():
            waypoint = ET.SubElement(flow_shape, "omgdi:waypoint", {"x": str(x), "y": str(y)})

    for node in bpmn_graph.get_nodes():
        process = process_process[node.get_process()]

        if isinstance(node, BPMN.StartEvent):
            isInterrupting = "true" if node.get_isInterrupting() else "false"
            parallelMultiple = "true" if node.get_parallelMultiple() else "false"
            task = ET.SubElement(process, "bpmn:startEvent",
                                 {"id": node.get_id(), "isInterrupting": isInterrupting, "name": node.get_name(),
                                  "parallelMultiple": parallelMultiple})
        elif isinstance(node, BPMN.EndEvent):
            task = ET.SubElement(process, "bpmn:endEvent", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.IntermediateCatchEvent):
            task = ET.SubElement(process, "bpmn:intermediateCatchEvent", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.IntermediateThrowEvent):
            task = ET.SubElement(process, "bpmn:intermediateThrowEvent", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.BoundaryEvent):
            task = ET.SubElement(process, "bpmn:boundaryEvent", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.Task):
            task = ET.SubElement(process, "bpmn:task", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.SubProcess):
            task = ET.SubElement(process, "bpmn:subProcess", {"id": node.get_id(), "name": node.get_name()})
        elif isinstance(node, BPMN.ExclusiveGateway):
            task = ET.SubElement(process, "bpmn:exclusiveGateway",
                                 {"id": node.get_id(), "gatewayDirection": node.get_gateway_direction().value.lower(),
                                  "name": ""})
        elif isinstance(node, BPMN.ParallelGateway):
            task = ET.SubElement(process, "bpmn:parallelGateway",
                                 {"id": node.get_id(), "gatewayDirection": node.get_gateway_direction().value.lower(),
                                  "name": ""})
        elif isinstance(node, BPMN.InclusiveGateway):
            task = ET.SubElement(process, "bpmn:inclusiveGateway",
                                 {"id": node.get_id(), "gatewayDirection": node.get_gateway_direction().value.lower(),
                                  "name": ""})
        else:
            raise Exception("Unexpected node type.")

        for in_arc in node.get_in_arcs():
            arc_xml = ET.SubElement(task, "bpmn:incoming")
            arc_xml.text = "id" + str(in_arc.get_id())

        for out_arc in node.get_out_arcs():
            arc_xml = ET.SubElement(task, "bpmn:outgoing")
            arc_xml.text = "id" + str(out_arc.get_id())

    for flow in bpmn_graph.get_flows():
        process = process_process[flow.get_process()]

        source = flow.get_source()
        target = flow.get_target()
        flow_xml = ET.SubElement(process, "bpmn:sequenceFlow", {"id": "id" + str(flow.get_id()), "name": flow.get_name(),
                                                           "sourceRef": str(source.get_id()),
                                                           "targetRef": str(target.get_id())})

    return minidom.parseString(ET.tostring(definitions)).toprettyxml(encoding=constants.DEFAULT_ENCODING)



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from collections import Counter
from itertools import product

from pm4py.algo.discovery.inductive.variants.im_clean import utils as im_utils
from pm4py.objects.log.obj import EventLog, Trace


def detect(alphabet, transitive_predecessors, transitive_successors):
    '''
    This method finds a xor cut in the dfg.
    Implementation follows function XorCut on page 188 of
    "Robust Process Mining with Guarantees" by Sander J.J. Leemans (ISBN: 978-90-386-4257-4)

    Basic Steps:
    1. create a group per activity
    2. merge pairwise reachable nodes (based on transitive relations)
    3. merge pairwise unreachable nodes (based on transitive relations)
    4. sort the groups based on their reachability
    
    Parameters
    ----------
    alphabet
        characters occurring in the dfg
    transitive_predecessors
        dictionary mapping activities to their (transitive) predecessors, according to the DFG
    transitive_successors
        dictionary mapping activities to their (transitive) successors, according to the DFG

    Returns
    -------
        A list of sets of activities, i.e., forming a maximal sequence cut
        None if no cut is found.

    '''
    groups = [{a} for a in alphabet]
    if len(groups) == 0:
        return None
    for a, b in product(alphabet, alphabet):
        if (b in transitive_successors[a] and a in transitive_successors[b]) or (
                b not in transitive_successors[a] and a not in transitive_successors[b]):
            groups = im_utils.__merge_groups_for_acts(a, b, groups)

    groups = list(sorted(groups, key=lambda g: len(
        transitive_predecessors[next(iter(g))]) + (len(alphabet) - len(transitive_successors[next(iter(g))]))))
    return groups if len(groups) > 1 else None


def project(log, groups, activity_key):
    '''
    This method projects the log based on a presumed sequence cut and a list of activity groups
    Parameters
    ----------
    log
        original log
    groups
        list of activity sets to be used in projection (activities can only appear in one group)
    activity_key
        key to use in the event to derive the activity name

    Returns
    -------
        list of corresponding logs according to the sequence cut.
    '''
    # refactored to support both IM and IMf
    logs = list()
    for group in groups:
        logs.append(EventLog())
    for t in log:
        i = 0
        split_point = 0
        act_union = set()
        while i < len(groups):
            new_split_point = find_split_point(t, groups[i], split_point, act_union, activity_key)
            trace_i = Trace()
            j = split_point
            while j < new_split_point:
                if t[j][activity_key] in groups[i]:
                    trace_i.append(t[j])
                j = j + 1
            logs[i].append(trace_i)
            split_point = new_split_point
            act_union = act_union.union(set(groups[i]))
            i = i + 1
    return logs


def find_split_point(t, group, start, ignore, activity_key):
    least_cost = 0
    position_with_least_cost = start
    cost = 0
    i = start
    while i < len(t):
        if t[i][activity_key] in group:
            cost = cost - 1
        elif t[i][activity_key] not in ignore:
            cost = cost + 1

        if cost < least_cost:
            least_cost = cost
            position_with_least_cost = i + 1

        i = i + 1

    return position_with_least_cost


def _is_strict_subset(A, B):
    return A != B and A.issubset(B)


def _is_strict_superset(A, B):
    return A != B and A.issuperset(B)


def project_dfg(dfg_sa_ea_actcount, groups):
    start_activities = []
    end_activities = []
    activities = []
    dfgs = []
    skippable = []
    for g in groups:
        skippable.append(False)
    activities_idx = {}
    for gind, g in enumerate(groups):
        for act in g:
            activities_idx[act] = int(gind)
    i = 0
    while i < len(groups):
        to_succ_arcs = Counter()
        from_prev_arcs = Counter()
        if i < len(groups) - 1:
            for arc in dfg_sa_ea_actcount.dfg:
                if arc[0] in groups[i] and arc[1] in groups[i + 1]:
                    to_succ_arcs[arc[0]] += dfg_sa_ea_actcount.dfg[arc]

        if i > 0:
            for arc in dfg_sa_ea_actcount.dfg:
                if arc[0] in groups[i - 1] and arc[1] in groups[i]:
                    from_prev_arcs[arc[1]] += dfg_sa_ea_actcount.dfg[arc]

        if i == 0:
            start_activities.append({})
            for act in dfg_sa_ea_actcount.start_activities:
                if act in groups[i]:
                    start_activities[i][act] = dfg_sa_ea_actcount.start_activities[act];
                else:
                    j = i
                    while j < activities_idx[act]:
                        skippable[j] = True
                        j = j + 1
        else:
            start_activities.append(from_prev_arcs)

        if i == len(groups) - 1:
            end_activities.append({})
            for act in dfg_sa_ea_actcount.end_activities:
                if act in groups[i]:
                    end_activities[i][act] = dfg_sa_ea_actcount.end_activities[act]
                else:
                    j = activities_idx[act] + 1
                    while j <= i:
                        skippable[j] = True
                        j = j + 1
        else:
            end_activities.append(to_succ_arcs)

        activities.append({})
        for act in groups[i]:
            activities[i][act] = dfg_sa_ea_actcount.act_count[act]
        dfgs.append({})
        for arc in dfg_sa_ea_actcount.dfg:
            if arc[0] in groups[i] and arc[1] in groups[i]:
                dfgs[i][arc] = dfg_sa_ea_actcount.dfg[arc]

        i = i + 1

    i = 0
    while i < len(dfgs):
        dfgs[i] = im_utils.DfgSaEaActCount(dfgs[i], start_activities[i], end_activities[i], activities[i])
        i = i + 1
    for arc in dfg_sa_ea_actcount.dfg:
        z = activities_idx[arc[1]]
        j = activities_idx[arc[0]] + 1
        while j < z:
            skippable[j] = False
            j = j + 1
    return [dfgs, skippable]



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import copy
from typing import List, Optional, Set, Dict

from pm4py.algo.discovery.inductive.variants.im_clean import utils as im_utils
from pm4py.algo.discovery.inductive.variants.im_clean.d_types import Cut, DFG
from pm4py.objects.log.obj import EventLog, Trace


def detect(dfg: DFG, alphabet: Dict[str, int], start_activities: Dict[str, int],
           end_activities: Dict[str, int]) -> Optional[Cut]:
    '''
    This method finds a loop cut in the dfg.
    Implementation follows function LoopCut on page 190 of
    "Robust Process Mining with Guarantees" by Sander J.J. Leemans (ISBN: 978-90-386-4257-4)

    Basic Steps:
    1. merge all start and end activities in one group ('do' group)
    2. remove start/end activities from the dfg
    3. detect connected components in (undirected representative) of the reduced graph
    4. check if each component meets the start/end criteria of the loop cut definition (merge with the 'do' group if not)
    5. return the cut if at least two groups remain

    Parameters
    ----------
    dfg
        directly follows graph
    alphabet
        alphabet of the dfg / log
    start_activities
        multiset of start activities of the dfg / log
    end_activities
        multiset of end activities of the dfg / log

    Returns
    -------
        A list of sets of activities, i.e., forming a maximal loop cut
        None if no cut is found.

    '''
    if len(dfg) == 0:
        return None

    groups = [set(start_activities.keys()).union(set(end_activities.keys()))]
    for c in _compute_connected_components(dfg, alphabet, start_activities, end_activities, groups[0]):
        groups.append(set(c.nodes))

    groups = _exclude_sets_non_reachable_from_start(dfg, start_activities, end_activities, groups)
    groups = _exclude_sets_no_reachable_from_end(dfg, start_activities, end_activities, groups)
    groups = _check_start_completeness(dfg, start_activities, end_activities, groups)
    groups = _check_end_completeness(dfg, start_activities, end_activities, groups)

    groups = list(filter(lambda g: len(g) > 0, groups))

    return groups if len(groups) > 1 else None


def _check_start_completeness(dfg, start_activities, end_activities, groups):
    i = 1
    while i < len(groups):
        merge = False
        for a in groups[i]:
            if merge:
                break
            for (x, b) in dfg:
                if x == a and b in start_activities:
                    for s in start_activities:
                        if not (a, s) in dfg:
                            merge = True
        if merge:
            groups[0] = groups[0].union(groups[i])
            del groups[i]
            continue
        i = i + 1
    return groups


def _check_end_completeness(dfg, start_activities, end_activities, groups):
    i = 1
    while i < len(groups):
        merge = False
        for a in groups[i]:
            if merge:
                break
            for (b, x) in dfg:
                if x == a and b in end_activities:
                    for e in end_activities:
                        if not (e, a) in dfg:
                            merge = True
        if merge:
            groups[0] = groups[0].union(groups[i])
            del groups[i]
            continue
        i = i + 1
    return groups


def _exclude_sets_non_reachable_from_start(dfg: DFG, start_activities: Dict[str, int], end_activities: Dict[str, int],
                                           groups: Cut) -> Cut:
    for a in set(start_activities).difference(set(end_activities)):
        for (x, b) in dfg:
            if x == a:
                group_a, group_b = None, None
                for group in groups:
                    group_a = group if a in group else group_a
                    group_b = group if b in group else group_b
                groups = [group for group in groups if group != group_a and group != group_b]
                # we are always merging on the do-part
                groups.insert(0, group_a.union(group_b))
    return groups


def _exclude_sets_no_reachable_from_end(dfg: DFG, start_activities: Dict[str, int], end_activities: Dict[str, int],
                                        groups: Cut) -> Cut:
    for b in set(end_activities).difference(set(start_activities)):
        for (a, x) in dfg:
            if x == b:
                group_a, group_b = None, None
                for group in groups:
                    group_a = group if a in group else group_a
                    group_b = group if b in group else group_b
                groups = [group for group in groups if group != group_a and group != group_b]
                groups.insert(0, group_a.union(group_b))
    return groups


def _compute_connected_components(dfg: DFG, alphabet: Set[str], start_activities: Dict[str, int],
                                  end_activities: Dict[str, int], do_set: Set[str]):
    import networkx as nx
    reduced_dfg = copy.copy(dfg)
    for (a, b) in dfg:
        if a in start_activities or a in end_activities or b in start_activities or b in end_activities:
            del reduced_dfg[(a, b)]
    reduced_alphabet = set(alphabet).difference(do_set)
    nx_directed = im_utils.transform_dfg_to_directed_nx_graph(reduced_dfg, reduced_alphabet)
    nx_undirected = nx_directed.to_undirected()
    return [nx_undirected.subgraph(c).copy() for c in nx.connected_components(nx_undirected)]


def project(log: EventLog, cut: Cut, activity_key: str) -> List[EventLog]:
    do = cut[0]
    redo = cut[1:]
    redo_activities = [y for x in redo for y in x]
    do_log = EventLog()
    redo_logs = []
    for i in range(len(redo)):
        redo_logs.append(EventLog())
    for t in log:
        do_trace = Trace()
        redo_trace = Trace()
        for e in t:
            if e[activity_key] in do:
                do_trace.append(e)
                if len(redo_trace) > 0:
                    redo_logs = _append_trace_to_redo_log(redo_trace, redo_logs, redo, activity_key)
                    redo_trace = Trace()
            else:
                if e[activity_key] in redo_activities:
                    redo_trace.append(e)
                    if len(do_trace) > 0:
                        do_log.append(do_trace)
                        do_trace = Trace()
        if len(redo_trace) > 0:
            redo_logs = _append_trace_to_redo_log(redo_trace, redo_logs, redo, activity_key)
        do_log.append(do_trace)
    logs = [do_log]
    logs.extend(redo_logs)
    return logs


def _append_trace_to_redo_log(redo_trace: Trace, redo_logs: List[List[Trace]], redo_groups: List[Set[str]],
                              activity_key: str) -> List[List[Trace]]:
    activities = set(x[activity_key] for x in redo_trace)
    inte = [(i, len(activities.intersection(redo_groups[i]))) for i in range(len(redo_groups))]
    inte = sorted(inte, key=lambda x: (x[1], x[0]), reverse=True)
    redo_logs[inte[0][0]].append(redo_trace)
    return redo_logs


def project_dfg(dfg_sa_ea_actcount, groups):
    dfgs = []
    skippable = [False, False]
    for gind, g in enumerate(groups):
        activities = {}
        start_activities = {}
        end_activities = {}
        paths_frequency = {}
        for act in dfg_sa_ea_actcount.act_count:
            if act in g:
                activities[act] = dfg_sa_ea_actcount.act_count[act]
        for arc in dfg_sa_ea_actcount.dfg:
            if arc[0] in g and arc[1] in g:
                paths_frequency[arc] = dfg_sa_ea_actcount.dfg[arc]
            if arc[1] in dfg_sa_ea_actcount.start_activities and arc[0] in dfg_sa_ea_actcount.end_activities:
                skippable[1] = True
        if gind == 0:
            for act in dfg_sa_ea_actcount.start_activities:
                if act in g:
                    start_activities[act] = dfg_sa_ea_actcount.start_activities[act]
                else:
                    skippable[0] = True
            for act in dfg_sa_ea_actcount.end_activities:
                if act in g:
                    end_activities[act] = dfg_sa_ea_actcount.end_activities[act]
                else:
                    skippable[1] = True
        elif gind == 1:
            for act in g:
                start_activities[act] = 1
                end_activities[act] = 1
        dfgs.append(im_utils.DfgSaEaActCount(paths_frequency, start_activities, end_activities, activities))
    return [dfgs, skippable]



from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from pm4py.algo.simulation.playout.process_tree import algorithm as tree_playout
import os


def execute_script():
    log = xes_importer.apply(os.path.join("..", "tests", "input_data", "running-example.xes"))
    tree = inductive_miner.apply_tree(log)
    new_log_1 = tree_playout.apply(tree)
    print(len(new_log_1))
    new_tree_1 = inductive_miner.apply_tree(new_log_1)
    print(new_tree_1)
    new_log_2 = tree_playout.apply(tree, variant=tree_playout.Variants.EXTENSIVE)
    print(len(new_log_2))
    new_tree_2 = inductive_miner.apply_tree(new_log_2)
    print(new_tree_2)


if __name__ == "__main__":
    execute_script()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum
from typing import Optional, Dict, Any, Union

from pm4py.objects.log.obj import EventLog, Trace
from pm4py.util import exec_utils, constants, xes_constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def apply(log: EventLog, act1: str, act2: str, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> EventLog:
    """
    Given an event log, filters all the subtraces going from an event with activity "act1" to an event with
    activity "act2"

    Parameters
    ----------------
    log
        Event log
    act1
        First activity
    act2
        Second activity
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key

    Returns
    ----------------
    filtered_log
        Log with all the subtraces going from "act1" to "act2"
    """
    if parameters is None:
        parameters = {}

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)

    filtered_log = EventLog(attributes=log.attributes, extensions=log.extensions, omni_present=log.omni_present,
                            classifiers=log.classifiers, properties=log.properties)

    for trace in log:
        act1_encountered = False
        filt_trace = None

        i = 0
        while i < len(trace) - 1:
            if not act1_encountered and trace[i][activity_key] == act1:
                act1_encountered = True
                filt_trace = Trace(attributes=trace.attributes)
                filt_trace.append(trace[i])
            elif act1_encountered and trace[i][activity_key] == act2:
                filt_trace.append(trace[i])
                filtered_log.append(filt_trace)
                act1_encountered = False
                filt_trace = None
            elif filt_trace is not None:
                filt_trace.append(trace[i])

            i = i + 1

    return filtered_log



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import math
import os
import tempfile
import uuid
from datetime import datetime
from enum import Enum
from typing import Optional, Dict, Any, Union

import matplotlib as mpl
import matplotlib.cm as cm

from pm4py.util import exec_utils
from pm4py.util.colors import get_string_from_int_below_255


class Parameters(Enum):
    FORMAT = "format"
    ACT_DIVIDER_SPACE = "act_divider_space"
    DATE_DIVIDER_SPACE = "date_divider_space"
    OVERALL_LENGTH_X = "overall_length_x"
    N_DIV_DATES = "n_div_dates"
    PERC_PATHS = "perc_paths"


def give_color_to_line(dir: float) -> str:
    """
    Gives a gradient color to the line

    Parameters
    ----------------
    dir
        Intensity of the difference (number between 0 and 1; 0=min difference, 1=max difference)

    Returns
    ----------------
    color
        Gradient color
    """
    dir = 0.5 + 0.5 * dir
    norm = mpl.colors.Normalize(vmin=0, vmax=1)
    nodes = [0.0, 0.01, 0.25, 0.4, 0.45, 0.55, 0.75, 0.99, 1.0]
    colors = ["deepskyblue", "skyblue", "lightcyan", "lightgray", "gray", "lightgray", "mistyrose", "salmon", "tomato"]
    cmap = mpl.colors.LinearSegmentedColormap.from_list("mycmap2", list(zip(nodes, colors)))
    # cmap = cm.plasma
    m = cm.ScalarMappable(norm=norm, cmap=cmap)
    rgba = m.to_rgba(dir)
    r = get_string_from_int_below_255(math.ceil(rgba[0] * 255.0))
    g = get_string_from_int_below_255(math.ceil(rgba[1] * 255.0))
    b = get_string_from_int_below_255(math.ceil(rgba[2] * 255.0))
    return "#" + r + g + b


def apply(perf_spectrum: Dict[str, Any], parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> str:
    """
    Construct the performance spectrum visualization

    Parameters
    ----------------
    perf_spectrum
        Performance spectrum
    parameters
        Parameters of the algorithm, including:
        - Parameters.FORMAT => format of the output (svg, png, ...)
        - Parameters.ACT_DIVIDER_SPACE => space between the activities in the spectrum
        - Parameters.DATE_DIVIDER_SPACE => space between the lines and the dates
        - Parmaeters.OVERALL_LENGTH_X => length of the X-line
        - Parameters.N_DIV_DATES => specifies the number of intermediate dates reported
        - Parameters.PERC_PATHS => (if provided) filter the (overall) most long paths

    Returns
    ---------------
    file_path
        Path containing the visualization
    """
    if parameters is None:
        parameters = {}

    format = exec_utils.get_param_value(Parameters.FORMAT, parameters, "png")
    act_divider = exec_utils.get_param_value(Parameters.ACT_DIVIDER_SPACE, parameters, 3.0)
    date_divider = exec_utils.get_param_value(Parameters.DATE_DIVIDER_SPACE, parameters, 1.0)
    overall_length = exec_utils.get_param_value(Parameters.OVERALL_LENGTH_X, parameters, 10.0)
    n_div = exec_utils.get_param_value(Parameters.N_DIV_DATES, parameters, 2)
    perc_paths = exec_utils.get_param_value(Parameters.PERC_PATHS, parameters, 1.0)

    output_file_gv = tempfile.NamedTemporaryFile(suffix=".gv")
    output_file_gv.close()

    output_file_img = tempfile.NamedTemporaryFile(suffix="." + format)
    output_file_img.close()

    lines = []
    lines.append("graph G {")

    min_x = min(x[0] for x in perf_spectrum["points"])
    max_x = max(x[-1] for x in perf_spectrum["points"])
    all_diffs = [x[i + 1] - x[i] for x in perf_spectrum["points"] for i in range(len(x) - 1)]
    min_diff = min(all_diffs)
    max_diff = max(all_diffs)

    points = sorted(perf_spectrum["points"], key=lambda x: x[-1] - x[0], reverse=True)
    points = points[:math.ceil(perc_paths * len(points))]

    for polyline in points:
        this_pts = []
        for i, p in enumerate(polyline):
            p_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
            first_coord = (p - min_x) / (max_x - min_x) * overall_length
            second_coord = act_divider * (len(perf_spectrum["list_activities"]) - i - 1)
            lines.append(
                "%s [label=\"\", pos=\"%.10f,%.10f!\", shape=none, width=\"0px\", height=\"0px\"];" % (p_id, first_coord, second_coord))
            this_pts.append(p_id)
        for i in range(len(this_pts) - 1):
            diff = polyline[i + 1] - polyline[i]
            color = give_color_to_line((diff - min_diff) / (max_diff - min_diff))
            lines.append("%s -- %s [ color=\"%s\" ];" % (this_pts[i], this_pts[i + 1], color))

    for i, act in enumerate(perf_spectrum["list_activities"]):
        second_coord = act_divider * (len(perf_spectrum["list_activities"]) - i - 1)
        a_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
        lines.append("%s [label=\"%s\", pos=\"%.10f,%.10f!\", shape=none, width=\"0px\", height=\"0px\"];" % (
        a_id, act, overall_length, second_coord))
        s_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
        lines.append("%s [label=\"\", pos=\"0,%.10f!\", shape=none, width=\"0px\", height=\"0px\"];" % (s_id, second_coord))
        lines.append("%s -- %s [ color=\"black\" ];" % (s_id, a_id))
        if i == len(perf_spectrum["list_activities"]) - 1:
            for j in range(n_div + 1):
                pos = float(j * overall_length) / float(n_div)
                tst = min_x + float(j) / float(n_div) * (max_x - min_x)
                dt = datetime.fromtimestamp(tst)
                n_id = "n" + str(uuid.uuid4()).replace("-", "") + "e"
                lines.append("%s [label=\"%s\", pos=\"%.10f,%.10f!\", shape=none, width=\"0px\", height=\"0px\"];" % (
                n_id, str(dt), pos, second_coord - date_divider))

    lines.append("}")
    lines = "\n".join(lines)

    F = open(output_file_gv.name, "w")
    F.write(lines)
    F.close()

    os.system("neato -T" + format + " " + output_file_gv.name + " > " + output_file_img.name)

    return output_file_img.name



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from enum import Enum

import pm4py
from pm4py.algo.discovery.dfg import algorithm as discover_dfg
from pm4py.algo.discovery.inductive.variants.im_clean.cuts import sequence as sequence_cut, xor as xor_cut, \
    concurrency as concurrent_cut, loop as loop_cut
from pm4py.algo.discovery.inductive.variants.im_clean.fall_throughs import activity_once_per_trace, activity_concurrent, \
    strict_tau_loop, tau_loop
from pm4py.algo.discovery.inductive.variants.im_clean.utils import __filter_dfg_on_threshold, __flower
from pm4py.algo.discovery.minimum_self_distance import algorithm as msd_algo
from pm4py.algo.discovery.minimum_self_distance import utils as msdw_algo
from pm4py.objects.dfg.utils import dfg_utils
from pm4py.objects.log.obj import EventLog
from pm4py.objects.log.util import log as log_util
from pm4py.objects.process_tree import obj as pt
from pm4py.statistics.end_activities.log import get as get_ends
from pm4py.statistics.start_activities.log import get as get_starters
from pm4py.util import constants


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY


def __inductive_miner(log, dfg, threshold, root, act_key, use_msd, remove_noise=False):
    tree = __inductive_miner_internal(log, dfg, threshold, root, act_key, use_msd, remove_noise)
    return tree


def __inductive_miner_internal(log, dfg, threshold, root, act_key, use_msd, remove_noise=False):
    alphabet = pm4py.get_event_attribute_values(log, act_key)
    if threshold > 0 and remove_noise:
        end_activities = get_ends.get_end_activities(log,
                                                     parameters={constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key})

        dfg = __filter_dfg_on_threshold(dfg, end_activities, threshold)

    original_length = len(log)
    log = pm4py.filter_log(lambda t: len(t) > 0, log)

    # revised EMPTYSTRACES
    if original_length - len(log) > original_length * threshold:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.XOR, root), threshold, act_key,
                                             [EventLog(), log],
                                             use_msd)

    start_activities = get_starters.get_start_activities(log, parameters={
        constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key})
    end_activities = get_ends.get_end_activities(log, parameters={constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key})

    if __is_base_case_act(log, act_key) or __is_base_case_silent(log):
        return __apply_base_case(log, root, act_key)
    pre, post = dfg_utils.get_transitive_relations(dfg, alphabet)
    cut = sequence_cut.detect(alphabet, pre, post)
    if cut is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.SEQUENCE, root), threshold, act_key,
                                             sequence_cut.project(log, cut, act_key), use_msd)
    cut = xor_cut.detect(dfg, alphabet)
    if cut is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.XOR, root), threshold, act_key,
                                             xor_cut.project(log, cut, act_key), use_msd)
    cut = concurrent_cut.detect(dfg, alphabet, start_activities, end_activities,
                                msd=msdw_algo.derive_msd_witnesses(log, msd_algo.apply(log, parameters={
                                    constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key}), parameters={
                                    constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key}) if use_msd else None)
    if cut is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.PARALLEL, root), threshold, act_key,
                                             concurrent_cut.project(log, cut, act_key), use_msd)
    cut = loop_cut.detect(dfg, alphabet, start_activities, end_activities)
    if cut is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.LOOP, root), threshold, act_key,
                                             loop_cut.project(log, cut, act_key), use_msd)

    aopt = activity_once_per_trace.detect(log, alphabet, act_key)
    if aopt is not None:
        operator = pt.ProcessTree(operator=pt.Operator.PARALLEL, parent=root)
        operator.children.append(pt.ProcessTree(operator=None, parent=operator, label=aopt))
        return __add_operator_recursive_logs(operator, threshold, act_key,
                                             activity_once_per_trace.project(log, aopt, act_key), use_msd)
    act_conc = activity_concurrent.detect(log, alphabet, act_key, use_msd)
    if act_conc is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.PARALLEL, root), threshold, act_key,
                                             activity_concurrent.project(log, act_conc, act_key), use_msd)
    stl = strict_tau_loop.detect(log, start_activities, end_activities, act_key)
    if stl is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.LOOP, root), threshold, act_key,
                                             [stl, EventLog()],
                                             use_msd)
    tl = tau_loop.detect(log, start_activities, act_key)
    if tl is not None:
        return __add_operator_recursive_logs(pt.ProcessTree(pt.Operator.LOOP, root), threshold, act_key,
                                             [tl, EventLog()],
                                             use_msd)

    if threshold > 0 and not remove_noise:
        return __inductive_miner(log, dfg, threshold, root, act_key, use_msd, remove_noise=True)

    return __flower(alphabet, root)


def __add_operator_recursive_logs(operator, threshold, act_key, logs, use_msd):
    if operator.operator != pt.Operator.LOOP:
        for log in logs:
            operator.children.append(__inductive_miner(log, discover_dfg.apply(log, parameters={
                constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key}), threshold, operator, act_key, use_msd))
    else:
        operator.children.append(__inductive_miner(logs[0], discover_dfg.apply(logs[0], parameters={
            constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key}), threshold, operator, act_key, use_msd))
        logs = logs[1:]
        if len(logs) == 1:
            operator.children.append(__inductive_miner(logs[0], discover_dfg.apply(logs[0], parameters={
                constants.PARAMETER_CONSTANT_ACTIVITY_KEY: act_key}), threshold, operator, act_key, use_msd))
        else:
            operator.children.append(
                __add_operator_recursive_logs(
                    pt.ProcessTree(operator=pt.Operator.XOR, parent=operator), threshold, act_key, logs, use_msd))
    return operator


def __is_base_case_act(log, act_key):
    if len(list(filter(lambda t: len(t) == 1, log))) == len(log):
        if len(frozenset(log_util.get_event_labels(log, act_key))) == 1:
            return True
    return False


def __is_base_case_silent(log):
    return len(log) == 0


def __apply_base_case(log, root, act_key):
    if len(log) == 0:
        operator = pt.ProcessTree(parent=root)
        return operator
    else:
        operator = pt.ProcessTree(parent=root, label=log[0][0][act_key])
        return operator



import inspect
import os
import sys
import traceback


def feature_extraction_ocel():
    from examples import feature_extraction_ocel
    print("\n\nfeature_extraction_ocel")
    feature_extraction_ocel.execute_script()


def ocel_validation():
    from examples import ocel_validation
    print("\n\nocel_validation")
    ocel_validation.execute_script()


def process_tree_frequency_annotation():
    from examples import process_tree_frequency_annotation
    print("\n\nprocess_tree_frequency_annotation")
    process_tree_frequency_annotation.execute_script()


def tree_manual_generation():
    from examples import tree_manual_generation
    print("\n\ntree_manual_generation")
    tree_manual_generation.execute_script()


def workalendar_example():
    from examples import workalendar_example
    print("\n\nworkalendar_example")
    workalendar_example.execute_script()


def merging_case_relations():
    from examples import merging_case_relations
    print("\n\nmerging_case_relations")
    merging_case_relations.execute_script()


def activity_position():
    from examples import activity_position
    print("\n\nactivity_position")
    activity_position.execute_script()


def link_analysis_vbfa():
    from examples import link_analysis_vbfa
    print("\n\nlink_analysis_vbfa")
    link_analysis_vbfa.execute_script()


def ocel_streaming():
    from examples import ocel_streaming
    print("\n\nocel_streaming")
    ocel_streaming.execute_script()


def petri_manual_generation():
    from examples import petri_manual_generation
    print("\n\npetri_manual_generation")
    petri_manual_generation.execute_script()


def timestamp_interleavings():
    from examples import timestamp_interleavings
    print("\n\ntimestamp_interleavings")
    timestamp_interleavings.execute_script()


def object_centric_petri_net_discovery():
    from examples import object_centric_petri_net_discovery
    print("\n\nobject_centric_petri_net_discovery")
    object_centric_petri_net_discovery.execute_script()


def trans_system_stochastic_view():
    from examples import trans_system_stochastic_vis
    print("\n\ntrans_system_stochastic_view")
    trans_system_stochastic_vis.execute_script()


def network_analysis():
    from examples import network_analysis
    print("\n\nnetwork_analysis")
    network_analysis.execute_script()


def read_write_ocel():
    from examples import read_write_ocel
    print("\n\nread_write_ocel")
    read_write_ocel.execute_script()


def ocdfg_discovery():
    from examples import ocdfg_discovery
    print("\n\nocdfg_discovery")
    ocdfg_discovery.execute_script()


def enrich_log_with_align():
    from examples import enrich_log_with_align
    print("\n\nenrich_log_with_align")
    enrich_log_with_align.execute_script()


def extended_marking_equation():
    from examples import extended_marking_equation
    print("\n\nextended_marking_equation")
    extended_marking_equation.execute_script()


def features_locally_linear_embedding():
    from examples import features_locally_linear_embedding
    print("\n\nfeatures_locally_linear_embedding")
    features_locally_linear_embedding.execute_script()


def discovery_data_petri_net():
    from examples import discovery_data_petri_net
    print("\n\ndiscovery_data_petri_net")
    discovery_data_petri_net.execute_script()


def pandas_iterable():
    from examples import pandas_iterable
    print("\n\npandas_iterable")
    pandas_iterable.execute_script()


def pandas_iterable_to_trace_stream():
    from examples import pandas_iterable_to_trace_stream
    print("\n\npandas_iterable_to_trace_stream")
    pandas_iterable_to_trace_stream.execute_script()


def performance_dfg_simulation():
    from examples import performance_dfg_simulation
    print("\n\nperformance_dfg_simulation")
    performance_dfg_simulation.execute_script()


def data_petri_nets():
    from examples import data_petri_nets
    print("\n\ndata_petri_nets")
    data_petri_nets.execute_script()


def inhibitor_reset_arcs():
    from examples import inhibitor_reset_arcs
    print("\n\ninhibitor_reset_arcs")
    inhibitor_reset_arcs.execute_script()


def batch_detection():
    from examples import batch_detection
    print("\n\nbatch_detection")
    batch_detection.execute_script()


def case_overlap_stat():
    from examples import case_overlap_stat
    print("\n\ncase_overlap_stat")
    case_overlap_stat.execute_script()


def cycle_time():
    from examples import cycle_time
    print("\n\ncycle_time")
    cycle_time.execute_script()


def rework():
    from examples import rework
    print("\n\nrework")
    rework.execute_script()


def events_distribution():
    from examples import events_distribution
    print("\n\nevents_distribution")
    events_distribution.execute_script()


def dotted_chart():
    from examples import dotted_chart
    print("\n\ndotted_chart")
    dotted_chart.execute_script()


def performance_spectrum():
    from examples import perf_spectrum_visualization
    print("\n\nperformance_spectrum")
    perf_spectrum_visualization.execute_script()


def woflan():
    from examples import woflan
    print("\n\nwoflan")
    woflan.execute_script()


def bpmn_from_pt():
    from examples import bpmn_from_pt_conversion
    print("\n\nbpmn_from_pt_conversion")
    bpmn_from_pt_conversion.execute_script()


def bpmn_import_and_to_petri_net():
    from examples import bpmn_import_and_to_petri_net
    print("\n\nbpmn_import_and_to_petri_net")
    bpmn_import_and_to_petri_net.execute_script()


def tree_playout():
    from examples import tree_playout
    print("\n\ntree_playout")
    tree_playout.execute_script()


def emd_evaluation():
    from examples import emd_evaluation
    print("\n\nemd_evaluation")
    emd_evaluation.execute_script()


def footprints_tree_conf():
    from examples import footprints_tree_conf
    print("\n\nfootprints_tree_conf")
    footprints_tree_conf.execute_script()


def simplified_interface():
    from examples import simplified_interface
    print("\n\nsimplified_interface")
    simplified_interface.execute_script()


def footprints_petri_net():
    from examples import footprints_petri_net
    print("\n\nfootprints_petri_net")
    footprints_petri_net.execute_script()


def corr_mining():
    from examples import corr_mining
    print("\n\ncorr_mining")
    corr_mining.execute_script()


def log_skeleton():
    from examples import log_skeleton
    print("\n\nlog_skeleton")
    log_skeleton.execute_script()


def roles_detection():
    from examples import roles_detection
    print("\n\nroles_detection")
    roles_detection.execute_script()


def alignment_test():
    from examples import alignment_test
    print("\n\nalignment_test")
    alignment_test.execute_script()


def dec_treplay_imdf():
    from examples import dec_treplay_imdf
    print("\n\ndec_treplay_imdf")
    dec_treplay_imdf.execute_script()


def logs_petri_visual_comparison():
    from examples import logs_petri_visual_comparison
    print("\n\nlogs_petri_visual_comparison")
    logs_petri_visual_comparison.execute_script()


def imdf_example():
    from examples import im_example
    print("\n\nimdf_example")
    im_example.execute_script()


def test_evaluation():
    from examples import test_evaluation
    print("\n\ntestEvaluation")
    test_evaluation.execute_script()


def sna_log():
    from examples import sna_log
    print("\n\nsna_log")
    sna_log.execute_script()


def token_replay_alpha():
    from examples import token_replay_alpha
    print("\n\ntokenReplay_alpha")
    token_replay_alpha.execute_script()


def manual_log_generation():
    from examples import manual_log_generation
    print("\n\nmanual_log_generation")
    manual_log_generation.execute_script()


def token_replay_imdf():
    from examples import token_replay_imdf
    print("\n\ntokenReplay_imdf")
    token_replay_imdf.execute_script()


def decisiontree_trivial_example():
    from examples import decisiontree_trivial_example
    print("\n\ndecisiontree_trivial_example")
    decisiontree_trivial_example.execute_script()


def decisiontree_align_example():
    from examples import decisiontree_align_example
    print("\n\ndecisiontree_align_example")
    decisiontree_align_example.execute_script()


def align_decomposition_example():
    from examples import align_decomposition_example
    print("\n\nalign_decomposition_example")
    align_decomposition_example.execute_script()


def transition_system_test():
    from examples import transition_system_test
    print("\n\ntransition_system_test")
    transition_system_test.execute_script()


def heu_miner_test():
    from examples import heu_miner_test
    print("\n\nheu_miner_test")
    heu_miner_test.execute_script()


def dfg_min_ex_log():
    from examples import dfg_min_ex_log
    print("\n\ndfg_min_ex_log")
    dfg_min_ex_log.execute_script()


def dfg_min_ex_pandas():
    from examples import dfg_min_ex_pandas
    print("\n\ndfg_min_ex_pandas")
    dfg_min_ex_pandas.execute_script()


def dfg_filt_act_paths_perc():
    from examples import dfg_filt_act_paths_perc
    print("\n\ndfg_filt_act_paths_perc")
    dfg_filt_act_paths_perc.execute_script()


def graphs_visualization():
    from examples import graphs_visualization
    print("\n\ngraphs_visualization")
    graphs_visualization.execute_script()


def backwards_token_replay():
    from examples import backwards_token_replay
    print("\n\nbackwards_token_replay")
    backwards_token_replay.execute_script()


def transient_dfg():
    from examples import transient_dfg
    print("\n\ntransient_dfg")
    transient_dfg.execute_script()


def transient_petri_net():
    from examples import transient_petri_net
    print("\n\ntransient_petri_net")
    transient_petri_net.execute_script()


def monte_carlo_dfg():
    from examples import montecarlo_dfg
    print("\n\nmontecarlo_dfg")
    montecarlo_dfg.execute_script()


def monte_carlo_petri_net():
    from examples import montecarlo_petri_net
    print("\n\nmontecarlo_petri_net")
    montecarlo_petri_net.execute_script()


def visualization_processtree():
    from examples import visualization_processtree
    print("\n\nvisualization_processtree")
    visualization_processtree.execute_script()


def diagn_add_dataframe():
    from examples import diagn_add_dataframe
    print("\n\ndiagn_add_dataframe")
    diagn_add_dataframe.execute_script()


def pn_to_pt():
    from examples import pn_to_pt
    print("\n\npn_to_pt")
    pn_to_pt.execute_script()


def visualization_align_table():
    from examples import visualization_align_table
    print("\n\nvisualization_align_table")
    visualization_align_table.execute_script()


def align_approx_pt():
    from examples import align_approx_pt
    print("\n\nalign_approx_pt")
    align_approx_pt.execute_script()


def streaming_conformance_footprints():
    from examples import streaming_conformance_footprints
    print("\n\nstreaming_conformance_footprints")
    streaming_conformance_footprints.execute_script()


def streaming_conformance_tbr():
    from examples import streaming_conformance_tbr
    print("\n\nstreaming_conformance_tbr")
    streaming_conformance_tbr.execute_script()


def streaming_conformance_temporal_profile():
    from examples import streaming_conformance_temporal_profile
    print("\n\nstreaming_conformance_temporal_profile")
    streaming_conformance_temporal_profile.execute_script()


def streaming_csv_reader_event_stream():
    from examples import streaming_csv_reader_event_stream
    print("\n\nstreaming_csv_reader_event_stream")
    streaming_csv_reader_event_stream.execute_script()


def streaming_discovery_dfg():
    from examples import streaming_discovery_dfg
    print("\n\nstreaming_discovery_dfg")
    streaming_discovery_dfg.execute_script()


def streaming_xes_reader_event_stream():
    from examples import streaming_xes_reader_event_stream
    print("\n\nstreaming_xes_reader_event_stream")
    streaming_xes_reader_event_stream.execute_script()


def streaming_xes_reader_trace_stream():
    from examples import streaming_xes_reader_trace_stream
    print("\n\nstreaming_xes_reader_trace_stream")
    streaming_xes_reader_trace_stream.execute_script()


def example_check_fitness():
    from examples import example_check_fitness
    print("\n\nexample_check_fitness")
    example_check_fitness.execute_script()


def dfg_align_log():
    from examples import dfg_align_log
    print("\n\ndfg_align_log")
    dfg_align_log.execute_script()


def dfg_playout():
    from examples import dfg_playout
    print("\n\ndfg_playout")
    dfg_playout.execute_script()


def temporal_profile_log():
    from examples import temporal_profile_log
    print("\n\ntemporal_profile_log")
    temporal_profile_log.execute_script()


def temporal_profile_dataframe():
    from examples import temporal_profile_dataframe
    print("\n\ntemporal_profile_dataframe")
    temporal_profile_dataframe.execute_script()


def dataframe_prefix_and_fea_extraction():
    from examples import dataframe_prefix_and_fea_extraction
    print("\n\ndataframe_prefix_and_fea_extraction")
    dataframe_prefix_and_fea_extraction.execute_script()


def logs_alignments():
    from examples import logs_alignment
    print("\n\nlogs_alignment")
    logs_alignment.execute_script()


def visualization_serialization():
    from examples import visualization_serialization
    print("\n\visualization_serialization")
    visualization_serialization.execute_script()


def orgmining_local_diagn():
    from examples import orgmining_local_diagn
    print("\n\norgmining_local_diagn")
    orgmining_local_diagn.execute_script()


def resource_profiles_log():
    from examples import resource_profiles_log
    print("\n\nresource_profiles_log")
    resource_profiles_log.execute_script()


def resource_profile_pandas():
    from examples import resource_profiles_pandas
    print("\n\nresource_profile_pandas")
    resource_profiles_pandas.execute_script()


def process_tree_reduction():
    from examples import process_tree_reduction
    print("\n\nprocess_tree_reduction")
    process_tree_reduction.execute_script()


def execute_script(f):
    try:
        f()
    except:
        traceback.print_exc()
        input("\npress INPUT if you want to continue")


sys.path.insert(0, os.path.dirname(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe())))))

if __name__ == "__main__":
    execute_script(feature_extraction_ocel)
    execute_script(ocel_validation)
    execute_script(process_tree_frequency_annotation)
    execute_script(tree_manual_generation)
    execute_script(workalendar_example)
    execute_script(merging_case_relations)
    execute_script(activity_position)
    execute_script(link_analysis_vbfa)
    execute_script(ocel_streaming)
    execute_script(petri_manual_generation)
    execute_script(timestamp_interleavings)
    execute_script(object_centric_petri_net_discovery)
    execute_script(trans_system_stochastic_view)
    execute_script(network_analysis)
    execute_script(read_write_ocel)
    execute_script(ocdfg_discovery)
    execute_script(enrich_log_with_align)
    execute_script(extended_marking_equation)
    execute_script(features_locally_linear_embedding)
    execute_script(discovery_data_petri_net)
    execute_script(pandas_iterable)
    execute_script(pandas_iterable_to_trace_stream)
    execute_script(data_petri_nets)
    execute_script(inhibitor_reset_arcs)
    execute_script(dotted_chart)
    execute_script(performance_spectrum)
    execute_script(orgmining_local_diagn)
    execute_script(resource_profiles_log)
    execute_script(resource_profile_pandas)
    execute_script(process_tree_reduction)
    execute_script(dataframe_prefix_and_fea_extraction)
    execute_script(logs_alignments)
    execute_script(visualization_serialization)
    execute_script(woflan)
    execute_script(bpmn_from_pt)
    execute_script(bpmn_import_and_to_petri_net)
    execute_script(tree_playout)
    execute_script(emd_evaluation)
    execute_script(footprints_tree_conf)
    execute_script(footprints_petri_net)
    execute_script(corr_mining)
    execute_script(log_skeleton)
    execute_script(roles_detection)
    execute_script(alignment_test)
    execute_script(dec_treplay_imdf)
    execute_script(logs_petri_visual_comparison)
    execute_script(imdf_example)
    execute_script(test_evaluation)
    execute_script(sna_log)
    execute_script(token_replay_alpha)
    execute_script(manual_log_generation)
    execute_script(token_replay_imdf)
    execute_script(decisiontree_trivial_example)
    execute_script(decisiontree_align_example)
    execute_script(align_decomposition_example)
    execute_script(transition_system_test)
    execute_script(heu_miner_test)
    execute_script(dfg_min_ex_log)
    execute_script(dfg_min_ex_pandas)
    execute_script(graphs_visualization)
    execute_script(backwards_token_replay)
    execute_script(transient_dfg)
    execute_script(transient_petri_net)
    execute_script(monte_carlo_dfg)
    execute_script(monte_carlo_petri_net)
    execute_script(visualization_processtree)
    execute_script(visualization_align_table)
    execute_script(pn_to_pt)
    execute_script(simplified_interface)
    execute_script(align_approx_pt)
    execute_script(streaming_conformance_footprints)
    execute_script(streaming_conformance_tbr)
    execute_script(streaming_csv_reader_event_stream)
    execute_script(streaming_discovery_dfg)
    execute_script(streaming_xes_reader_event_stream)
    execute_script(streaming_xes_reader_trace_stream)
    execute_script(example_check_fitness)



import os
import unittest
import pandas as pd
from pm4py.algo.discovery.log_skeleton import algorithm as lsk_alg
from pm4py.algo.conformance.log_skeleton import algorithm as lsk_conf_alg
from pm4py.objects.process_tree.importer import importer as ptree_importer
from pm4py.objects.process_tree.exporter import exporter as ptree_exporter
from pm4py.algo.discovery.performance_spectrum.variants import log as log_pspectrum, dataframe as df_pspectrum
from pm4py.objects.dfg.importer import importer as dfg_importer
from pm4py.objects.dfg.exporter import exporter as dfg_exporter
from pm4py.algo.discovery.dfg import algorithm as dfg_discovery
from pm4py.statistics.start_activities.log import get as start_activities
from pm4py.statistics.end_activities.log import get as end_activities
from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
from pm4py.statistics.variants.log import get as variants_get
from pm4py.algo.simulation.playout.petri_net import algorithm
from pm4py.objects.conversion.log import converter
from pm4py.objects.log.util import dataframe_utils
from pm4py.util import pandas_utils


class OtherPartsTests(unittest.TestCase):
    def test_emd_1(self):
        from pm4py.algo.evaluation.earth_mover_distance import algorithm as earth_mover_distance
        M = {("a", "b", "d", "e"): 0.49, ("a", "d", "b", "e"): 0.49, ("a", "c", "d", "e"): 0.01,
             ("a", "d", "c", "e"): 0.01}
        L1 = {("a", "b", "d", "e"): 0.49, ("a", "d", "b", "e"): 0.49, ("a", "c", "d", "e"): 0.01,
              ("a", "d", "c", "e"): 0.01}
        earth_mover_distance.apply(M, L1)

    def test_emd_2(self):
        from pm4py.algo.evaluation.earth_mover_distance import algorithm as earth_mover_distance
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        lang_log = variants_get.get_language(log)
        net1, im1, fm1 = inductive_miner.apply(log)
        lang_model1 = variants_get.get_language(
            algorithm.apply(net1, im1, fm1, variant=algorithm.Variants.STOCHASTIC_PLAYOUT,
                            parameters={algorithm.Variants.STOCHASTIC_PLAYOUT.value.Parameters.LOG: log}))
        emd = earth_mover_distance.apply(lang_model1, lang_log)

    def test_importing_dfg(self):
        dfg, sa, ea = dfg_importer.apply(os.path.join("input_data", "running-example.dfg"))

    def test_exporting_dfg(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        dfg = dfg_discovery.apply(log)
        dfg_exporter.apply(dfg, os.path.join("test_output_data", "running-example.dfg"))
        dfg, sa, ea = dfg_importer.apply(os.path.join("test_output_data", "running-example.dfg"))
        os.remove(os.path.join("test_output_data", "running-example.dfg"))

    def test_exporting_dfg_with_sa_ea(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        dfg = dfg_discovery.apply(log)
        sa = start_activities.get_start_activities(log)
        ea = end_activities.get_end_activities(log)
        dfg_exporter.apply(dfg, os.path.join("test_output_data", "running-example.dfg"),
                           parameters={dfg_exporter.Variants.CLASSIC.value.Parameters.START_ACTIVITIES: sa,
                                       dfg_exporter.Variants.CLASSIC.value.Parameters.END_ACTIVITIES: ea})
        dfg, sa, ea = dfg_importer.apply(os.path.join("test_output_data", "running-example.dfg"))
        os.remove(os.path.join("test_output_data", "running-example.dfg"))

    def test_log_skeleton(self):
        log = xes_importer.apply(os.path.join("input_data", "receipt.xes"))
        skeleton = lsk_alg.apply(log)
        conf_res = lsk_conf_alg.apply(log, skeleton)

    def test_performance_spectrum_log(self):
        log = xes_importer.apply(os.path.join("input_data", "receipt.xes"))
        pspectr = log_pspectrum.apply(log, ["T02 Check confirmation of receipt", "T03 Adjust confirmation of receipt"],
                                      1000, {})

    def test_performance_spectrum_df(self):
        df = pd.read_csv(os.path.join("input_data", "receipt.csv"))
        df = dataframe_utils.convert_timestamp_columns_in_df(df)
        pspectr = df_pspectrum.apply(df, ["T02 Check confirmation of receipt", "T03 Adjust confirmation of receipt"],
                                     1000, {})

    def test_alignment(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.alpha import algorithm as alpha_miner
        net, im, fm = alpha_miner.apply(log)
        from pm4py.algo.conformance.alignments.petri_net import algorithm as alignments
        aligned_traces = alignments.apply(log, net, im, fm, variant=alignments.Variants.VERSION_STATE_EQUATION_A_STAR)
        aligned_traces = alignments.apply(log, net, im, fm, variant=alignments.Variants.VERSION_DIJKSTRA_NO_HEURISTICS)

    def test_import_export_ptml(self):
        tree = ptree_importer.apply(os.path.join("input_data", "running-example.ptml"))
        ptree_exporter.apply(tree, os.path.join("test_output_data", "running-example2.ptml"))
        os.remove(os.path.join("test_output_data", "running-example2.ptml"))

    def test_footprints_net(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.alpha import algorithm as alpha_miner
        net, im, fm = alpha_miner.apply(log)
        from pm4py.algo.discovery.footprints import algorithm as footprints_discovery
        fp_entire_log = footprints_discovery.apply(log, variant=footprints_discovery.Variants.ENTIRE_EVENT_LOG)
        fp_trace_trace = footprints_discovery.apply(log)
        fp_net = footprints_discovery.apply(net, im)
        from pm4py.algo.conformance.footprints import algorithm as footprints_conformance
        conf1 = footprints_conformance.apply(fp_entire_log, fp_net)
        conf2 = footprints_conformance.apply(fp_trace_trace, fp_net)
        conf3 = footprints_conformance.apply(fp_entire_log, fp_net,
                                             variant=footprints_conformance.Variants.LOG_EXTENSIVE)
        conf4 = footprints_conformance.apply(fp_trace_trace, fp_net,
                                             variant=footprints_conformance.Variants.TRACE_EXTENSIVE)

    def test_footprints_tree(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.inductive import algorithm as inductive_miner
        tree = inductive_miner.apply_tree(log)
        from pm4py.algo.discovery.footprints import algorithm as footprints_discovery
        fp_entire_log = footprints_discovery.apply(log, variant=footprints_discovery.Variants.ENTIRE_EVENT_LOG)
        fp_trace_trace = footprints_discovery.apply(log)
        fp_tree = footprints_discovery.apply(tree)
        from pm4py.algo.conformance.footprints import algorithm as footprints_conformance
        conf1 = footprints_conformance.apply(fp_entire_log, fp_tree)
        conf2 = footprints_conformance.apply(fp_trace_trace, fp_tree)
        conf3 = footprints_conformance.apply(fp_entire_log, fp_tree,
                                             variant=footprints_conformance.Variants.LOG_EXTENSIVE)
        conf4 = footprints_conformance.apply(fp_trace_trace, fp_tree,
                                             variant=footprints_conformance.Variants.TRACE_EXTENSIVE)

    def test_footprints_tree_df(self):
        df = pd.read_csv(os.path.join("input_data", "running-example.csv"))
        df = dataframe_utils.convert_timestamp_columns_in_df(df)
        from pm4py.algo.discovery.inductive import algorithm as inductive_miner
        log = converter.apply(df)
        tree = inductive_miner.apply_tree(log)
        from pm4py.algo.discovery.footprints import algorithm as footprints_discovery
        fp_df = footprints_discovery.apply(df)
        fp_tree = footprints_discovery.apply(tree)
        from pm4py.algo.conformance.footprints import algorithm as footprints_conformance
        conf = footprints_conformance.apply(fp_df, fp_tree)

    def test_conversion_pn_to_pt(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.alpha import algorithm as alpha_miner
        net, im, fm = alpha_miner.apply(log)
        from pm4py.objects.conversion.wf_net import converter as wf_net_converter
        tree = wf_net_converter.apply(net, im, fm, variant=wf_net_converter.Variants.TO_PROCESS_TREE)

    def test_playout_tree_basic(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.inductive import algorithm as inductive_miner
        tree = inductive_miner.apply_tree(log)
        from pm4py.algo.simulation.playout.process_tree import algorithm as tree_playout
        new_log = tree_playout.apply(tree)

    def test_playout_tree_extensive(self):
        log = xes_importer.apply(os.path.join("input_data", "running-example.xes"))
        from pm4py.algo.discovery.inductive import algorithm as inductive_miner
        tree = inductive_miner.apply_tree(log)
        from pm4py.algo.simulation.playout.process_tree import algorithm as tree_playout
        new_log = tree_playout.apply(tree, variant=tree_playout.Variants.EXTENSIVE)

    def test_sojourn_time_xes(self):
        log = xes_importer.apply(os.path.join("input_data", "interval_event_log.xes"))
        from pm4py.statistics.sojourn_time.log import get
        soj_time = get.apply(log, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_sojourn_time_pandas(self):
        import pandas as pd
        dataframe = pd.read_csv(os.path.join("input_data", "interval_event_log.csv"))
        from pm4py.objects.log.util import dataframe_utils
        dataframe = dataframe_utils.convert_timestamp_columns_in_df(dataframe)
        from pm4py.statistics.sojourn_time.pandas import get
        soj_time = get.apply(dataframe, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_concurrent_activities_xes(self):
        log = xes_importer.apply(os.path.join("input_data", "interval_event_log.xes"))
        from pm4py.statistics.concurrent_activities.log import get
        conc_act = get.apply(log, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_concurrent_activities_pandas(self):
        import pandas as pd
        dataframe = pd.read_csv(os.path.join("input_data", "interval_event_log.csv"))
        from pm4py.objects.log.util import dataframe_utils
        dataframe = dataframe_utils.convert_timestamp_columns_in_df(dataframe)
        from pm4py.statistics.concurrent_activities.pandas import get
        conc_act = get.apply(dataframe, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_efg_xes(self):
        log = xes_importer.apply(os.path.join("input_data", "interval_event_log.xes"))
        from pm4py.statistics.eventually_follows.log import get
        efg = get.apply(log, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_efg_pandas(self):
        import pandas as pd
        dataframe = pd.read_csv(os.path.join("input_data", "interval_event_log.csv"))
        from pm4py.objects.log.util import dataframe_utils
        dataframe = dataframe_utils.convert_timestamp_columns_in_df(dataframe)
        from pm4py.statistics.eventually_follows.pandas import get
        efg = get.apply(dataframe, parameters={get.Parameters.START_TIMESTAMP_KEY: "start_timestamp"})

    def test_dfg_playout(self):
        import pm4py
        from pm4py.algo.simulation.playout.dfg import algorithm as dfg_playout
        log = pm4py.read_xes(os.path.join("input_data", "running-example.xes"))
        dfg, sa, ea = pm4py.discover_dfg(log)
        dfg_playout.apply(dfg, sa, ea)

    def test_dfg_align(self):
        import pm4py
        from pm4py.algo.filtering.dfg import dfg_filtering
        from pm4py.algo.conformance.alignments.dfg import algorithm as dfg_alignment
        log = pm4py.read_xes(os.path.join("input_data", "running-example.xes"))
        dfg, sa, ea = pm4py.discover_dfg(log)
        act_count = pm4py.get_event_attribute_values(log, "concept:name")
        dfg, sa, ea, act_count = dfg_filtering.filter_dfg_on_activities_percentage(dfg, sa, ea, act_count, 0.5)
        dfg, sa, ea, act_count = dfg_filtering.filter_dfg_on_paths_percentage(dfg, sa, ea, act_count, 0.5)
        aligned_traces = dfg_alignment.apply(log, dfg, sa, ea)

    def test_insert_idx_in_trace(self):
        df = pd.read_csv(os.path.join("input_data", "running-example.csv"))
        df = pandas_utils.insert_ev_in_tr_index(df)

    def test_automatic_feature_extraction(self):
        df = pd.read_csv(os.path.join("input_data", "receipt.csv"))
        fea_df = dataframe_utils.automatic_feature_extraction_df(df)

    def test_log_to_trie(self):
        import pm4py
        from pm4py.algo.transformation.log_to_trie import algorithm as log_to_trie
        log = pm4py.read_xes(os.path.join("input_data", "running-example.xes"))
        trie = log_to_trie.apply(log)

    def test_minimum_self_distance(self):
        import pm4py
        from pm4py.algo.discovery.minimum_self_distance import algorithm as minimum_self_distance
        log = pm4py.read_xes(os.path.join("input_data", "running-example.xes"))
        msd = minimum_self_distance.apply(log)

    def test_lp_solver(self):
        import pm4py
        if "cvxopt" not in pm4py.util.lp.solver.DEFAULT_LP_SOLVER_VARIANT:
            raise Exception("cvxopt is not the solver")


if __name__ == "__main__":
    unittest.main()



from pm4py.objects.log.importer.xes import importer as xes_importer
from pm4py.algo.discovery.dfg import algorithm as dfg_miner
from pm4py.objects.conversion.dfg import converter as dfg_conv
from pm4py.algo.simulation.montecarlo import algorithm as montecarlo_simulation
from pm4py.algo.conformance.tokenreplay.algorithm import Variants
import os


def execute_script():
    log = xes_importer.apply(os.path.join("..", "tests", "input_data", "running-example.xes"))
    frequency_dfg = dfg_miner.apply(log, variant=dfg_miner.Variants.FREQUENCY)
    net, im, fm = dfg_conv.apply(frequency_dfg)
    # perform the Montecarlo simulation with the arrival rate inferred by the log (the simulation lasts 5 secs)
    parameters = {}
    parameters[
        montecarlo_simulation.Variants.PETRI_SEMAPH_FIFO.value.Parameters.TOKEN_REPLAY_VARIANT] = Variants.BACKWARDS
    parameters[montecarlo_simulation.Variants.PETRI_SEMAPH_FIFO.value.Parameters.PARAM_ENABLE_DIAGNOSTICS] = False
    parameters[montecarlo_simulation.Variants.PETRI_SEMAPH_FIFO.value.Parameters.PARAM_MAX_THREAD_EXECUTION_TIME] = 5
    log, res = montecarlo_simulation.apply(log, net, im, fm, parameters=parameters)
    print("\n(Montecarlo - Petri net) case arrival ratio inferred from the log")
    print(res["median_cases_ex_time"])
    print(res["total_cases_time"])
    # perform the Montecarlo simulation with the arrival rate specified (the simulation lasts 5 secs)
    parameters[montecarlo_simulation.Variants.PETRI_SEMAPH_FIFO.value.Parameters.PARAM_CASE_ARRIVAL_RATIO] = 60
    log, res = montecarlo_simulation.apply(log, net, im, fm, parameters=parameters)
    print("\n(Montecarlo - Petri net) case arrival ratio specified by the user")
    print(res["median_cases_ex_time"])
    print(res["total_cases_time"])


if __name__ == "__main__":
    execute_script()



import pm4py
from pm4py.objects.process_tree.obj import ProcessTree, Operator


def execute_script():
    root = ProcessTree(operator=Operator.SEQUENCE)

    choice = ProcessTree(operator=Operator.XOR, parent=root)
    parallel = ProcessTree(operator=Operator.PARALLEL, parent=root)

    root.children.append(choice)
    root.children.append(parallel)

    leaf_A = ProcessTree(label="A", parent=choice)
    leaf_B = ProcessTree(label="B", parent=choice)
    leaf_C = ProcessTree(label="C", parent=choice)

    choice.children.append(leaf_A)
    choice.children.append(leaf_B)
    choice.children.append(leaf_C)

    leaf_D = ProcessTree(label="D", parent=parallel)
    leaf_E = ProcessTree(label="E", parent=parallel)
    leaf_F = ProcessTree(label="F", parent=parallel)

    parallel.children.append(leaf_D)
    parallel.children.append(leaf_E)
    parallel.children.append(leaf_F)

    pm4py.view_process_tree(root, format="svg")

    # remove leaf_C from choice
    choice.children.remove(leaf_C)

    # remove the leaf with label "E" from parallel
    parallel.children.remove(
        [parallel.children[i] for i in range(len(parallel.children)) if parallel.children[i].label == "E"][0])

    pm4py.view_process_tree(root, format="svg")


if __name__ == "__main__":
    execute_script()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.petri_net.utils.reachability_graph import construct_reachability_graph
from pm4py.objects.conversion.log import converter as log_converter


def get_tangible_reachability_from_log_net_im_fm(log, net, im, fm, parameters=None):
    """
    Gets the tangible reachability graph from a log and an accepting Petri net

    Parameters
    ---------------
    log
        Event log
    net
        Petri net
    im
        Initial marking
    fm
        Final marking

    Returns
    ------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    stochastic_info
        Stochastic information
    """
    if parameters is None:
        parameters = {}

    from pm4py.algo.simulation.montecarlo.utils import replay
    stochastic_info = replay.get_map_from_log_and_net(log_converter.apply(log, parameters=parameters), net, im, fm,
                                                      parameters=parameters)

    reachability_graph, tangible_reachability_graph = get_tangible_reachability_from_net_im_sinfo(net, im,
                                                                                                  stochastic_info,
                                                                                                  parameters=parameters)

    return reachability_graph, tangible_reachability_graph, stochastic_info


def get_tangible_reachability_from_net_im_sinfo(net, im, stochastic_info, parameters=None):
    """
    Gets the tangible reacahbility graph from a Petri net, an initial marking and a stochastic map

    Parameters
    -------------
    net
        Petri net
    im
        Initial marking
    fm
        Final marking
    stochastic_info
        Stochastic information

    Returns
    ------------
    reachab_graph
        Reachability graph
    tangible_reach_graph
        Tangible reachability graph
    """
    if parameters is None:
        parameters = {}
    reachab_graph = construct_reachability_graph(net, im)
    tang_reach_graph = get_tangible_reachability_from_reachability(reachab_graph, stochastic_info)

    return reachab_graph, tang_reach_graph


def get_tangible_reachability_from_reachability(reach_graph, stochastic_info):
    """
    Gets the tangible reachability graph from the reachability graph and the stochastic transition map

    Parameters
    ------------
    reach_graph
        Reachability graph
    stochastic_info
        Stochastic information

    Returns
    ------------
    tangible_reach_graph
        Tangible reachability graph
    """
    timed_transitions = []
    for trans in stochastic_info.keys():
        random_variable = stochastic_info[trans]
        transition_type = random_variable.get_transition_type()
        if transition_type == "TIMED":
            timed_transitions.append(trans.name)
    states_reach = list(reach_graph.states)
    for s in states_reach:
        state_outgoing_trans = list(s.outgoing)
        state_ingoing_trans = list(s.incoming)
        timed_trans_outgoing = [x for x in state_outgoing_trans if x.name in timed_transitions]
        if not len(state_outgoing_trans) == len(timed_trans_outgoing):
            for t in state_outgoing_trans:
                reach_graph.transitions.remove(t)
                t.from_state.outgoing.remove(t)
                t.to_state.incoming.remove(t)
            for t in state_ingoing_trans:
                reach_graph.transitions.remove(t)
                t.from_state.outgoing.remove(t)
                t.to_state.incoming.remove(t)
            reach_graph.states.remove(s)
    states_reach = list(reach_graph.states)
    for s in states_reach:
        if len(s.incoming) == 0 and len(s.outgoing) == 0:
            reach_graph.states.remove(s)

    return reach_graph



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import pandas as pd
from enum import Enum

from pm4py.util import exec_utils, constants, xes_constants
from pm4py.util.business_hours import soj_time_business_hours_diff
from typing import Optional, Dict, Any, Union


class Parameters(Enum):
    ACTIVITY_KEY = constants.PARAMETER_CONSTANT_ACTIVITY_KEY
    START_TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_START_TIMESTAMP_KEY
    TIMESTAMP_KEY = constants.PARAMETER_CONSTANT_TIMESTAMP_KEY
    AGGREGATION_MEASURE = "aggregationMeasure"
    BUSINESS_HOURS = "business_hours"
    WORKTIMING  = "worktiming"
    WEEKENDS = "weekends"
    WORKCALENDAR = "workcalendar"


DIFF_KEY = "@@diff"


def apply(dataframe: pd.DataFrame, parameters: Optional[Dict[Union[str, Parameters], Any]] = None) -> Dict[str, float]:
    """
    Gets the sojourn time per activity on a Pandas dataframe

    Parameters
    --------------
    dataframe
        Pandas dataframe
    parameters
        Parameters of the algorithm, including:
        - Parameters.ACTIVITY_KEY => activity key
        - Parameters.START_TIMESTAMP_KEY => start timestamp key
        - Parameters.TIMESTAMP_KEY => timestamp key
        - Parameters.BUSINESS_HOURS => calculates the difference of time based on the business hours, not the total time.
                                        Default: False
        - Parameters.WORKTIMING => work schedule of the company (provided as a list where the first number is the start
            of the work time, and the second number is the end of the work time), if business hours are enabled
                                        Default: [7, 17] (work shift from 07:00 to 17:00)
        - Parameters.WEEKENDS => indexes of the days of the week that are weekend
                                        Default: [6, 7] (weekends are Saturday and Sunday)
        - Parameters.AGGREGATION_MEASURE => performance aggregation measure (sum, min, max, mean, median)

    Returns
    --------------
    soj_time_dict
        Sojourn time dictionary
    """
    if parameters is None:
        parameters = {}

    business_hours = exec_utils.get_param_value(Parameters.BUSINESS_HOURS, parameters, False)
    worktiming = exec_utils.get_param_value(Parameters.WORKTIMING, parameters, [7, 17])
    weekends = exec_utils.get_param_value(Parameters.WEEKENDS, parameters, [6, 7])
    workcalendar = exec_utils.get_param_value(Parameters.WORKCALENDAR, parameters, constants.DEFAULT_BUSINESS_HOURS_WORKCALENDAR)

    activity_key = exec_utils.get_param_value(Parameters.ACTIVITY_KEY, parameters, xes_constants.DEFAULT_NAME_KEY)
    start_timestamp_key = exec_utils.get_param_value(Parameters.START_TIMESTAMP_KEY, parameters,
                                                     xes_constants.DEFAULT_TIMESTAMP_KEY)
    timestamp_key = exec_utils.get_param_value(Parameters.TIMESTAMP_KEY, parameters,
                                               xes_constants.DEFAULT_TIMESTAMP_KEY)
    aggregation_measure = exec_utils.get_param_value(Parameters.AGGREGATION_MEASURE,
                                                     parameters, "mean")

    if business_hours:
        dataframe[DIFF_KEY] = dataframe.apply(
            lambda x: soj_time_business_hours_diff(x[start_timestamp_key], x[timestamp_key], worktiming,
                                                   weekends, workcalendar), axis=1)
    else:
        dataframe[DIFF_KEY] = (
            dataframe[timestamp_key] - dataframe[start_timestamp_key]
        ).astype('timedelta64[s]')

    dataframe = dataframe.reset_index()

    column = dataframe.groupby(activity_key)[DIFF_KEY]
    if aggregation_measure == "median":
        ret_dict = column.median().to_dict()
    elif aggregation_measure == "min":
        ret_dict = column.min().to_dict()
    elif aggregation_measure == "max":
        ret_dict = column.max().to_dict()
    elif aggregation_measure == "sum":
        ret_dict = column.sum().to_dict()
    else:
        ret_dict = column.mean().to_dict()

    # assure to avoid problems with np.float64, by using the Python float type
    for el in ret_dict:
        ret_dict[el] = float(ret_dict[el])

    return ret_dict



import time

import pm4py
from pm4py.algo.conformance.alignments.edit_distance import algorithm as logs_alignment
from pm4py.algo.evaluation.replay_fitness.variants import alignment_based
from pm4py.objects.process_tree.utils import bottomup as bottomup_discovery
from pm4py.algo.simulation.playout.process_tree import algorithm as tree_playout


def execute_script():
    # read an event log
    log = pm4py.read_xes("../tests/compressed_input_data/02_teleclaims.xes.gz")
    # log = pm4py.read_xes("../tests/input_data/receipt.xes")
    print("number of variants of the original log ->", len(pm4py.get_variants_as_tuples(log)))
    # discover a process model
    tree = pm4py.discover_process_tree_inductive(log)
    # simulate a log out of the model (to have another log that is similar to the original)
    aa = time.time()
    min_trace_length = bottomup_discovery.get_min_trace_length(tree)
    simulated_log = tree_playout.apply(tree, variant=tree_playout.Variants.EXTENSIVE,
                                       parameters={"max_trace_length": min_trace_length + 2})
    print("number of variants of the simulated log -> ", len(simulated_log))
    # apply the alignments between this log and the model
    bb = time.time()
    aligned_traces = logs_alignment.apply(log, simulated_log)
    cc = time.time()
    print(aligned_traces[0])
    print("playout time", bb - aa)
    print("alignments time", cc - bb)
    print("TOTAL", cc - aa)
    print(alignment_based.evaluate(aligned_traces))
    # apply the anti alignments between this log and the model
    dd = time.time()
    anti_aligned_traces = logs_alignment.apply(log, simulated_log,
                                               parameters={
                                                   logs_alignment.Variants.EDIT_DISTANCE.value.Parameters.PERFORM_ANTI_ALIGNMENT: True})
    ee = time.time()
    print(anti_aligned_traces[0])
    print("anti alignments time", ee - dd)
    print(alignment_based.evaluate(anti_aligned_traces))


if __name__ == "__main__":
    execute_script()



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
from pm4py.objects.ocel.obj import OCEL
from typing import Optional, Dict, Any
from pm4py.algo.transformation.ocel.features.objects import object_lifecycle_duration
import pandas as pd


def apply(ocel: OCEL, parameters: Optional[Dict[Any, Any]] = None):
    """
    Assigns to each object a feature which is the number of objects that are open during the lifecycle
    of the object.

    Parameters
    ----------------
    ocel
        OCEL
    parameters
        Parameters of the algorithm:
            - Parameters.OBJECT_STR_ATTRIBUTES => collection of string attributes to consider for feature extraction.

    Returns
    ----------------
    data
        Extracted feature values
    feature_names
        Feature names
    """
    if parameters is None:
        parameters = {}

    from intervaltree import Interval, IntervalTree
    
    data, feature_names = object_lifecycle_duration.apply(ocel, parameters=parameters)
    obj_dur = pd.DataFrame(data, columns=feature_names)
    obj_dur = obj_dur[["@@object_lifecycle_start_timestamp", "@@object_lifecycle_end_timestamp"]].to_dict("records")
    small_k = 10**-5

    tree = IntervalTree()

    for el in obj_dur:
        tree.add(Interval(el["@@object_lifecycle_start_timestamp"] - small_k,
                          el["@@object_lifecycle_end_timestamp"] + small_k))

    data = []
    feature_names = ["@@object_wip"]

    for el in obj_dur:
        data.append([len(tree[el["@@object_lifecycle_start_timestamp"] - small_k: el["@@object_lifecycle_end_timestamp"] + small_k])])

    return data, feature_names



'''
    This file is part of PM4Py (More Info: https://pm4py.fit.fraunhofer.de).

    PM4Py is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    PM4Py is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with PM4Py.  If not, see <https://www.gnu.org/licenses/>.
'''
import numpy as np
import sympy

def compute_place_invariants(net):
    """
    We compute the NUllspace of the incidence matrix and obtain the place-invariants.
    :param net: Petri Net of which we want to know the place invariants.
    :return: Set of place invariants of the given Petri Net.
    """

    def compute_incidence_matrix(net):
        """
        Given a Petri Net, the incidence matrix is computed. An incidence matrix has n rows (places) and m columns
        (transitions).
        :param net: Petri Net object
        :return: Incidence matrix
        """
        n = len(net.transitions)
        m = len(net.places)
        C = np.zeros((m, n))
        i = 0
        transition_list = list(net.transitions)
        place_list = list(net.places)
        while i < n:
            t = transition_list[i]
            for in_arc in t.in_arcs:
                # arcs that go to transition
                C[place_list.index(in_arc.source), i] -= 1
            for out_arc in t.out_arcs:
                # arcs that lead away from transition
                C[place_list.index(out_arc.target), i] += 1
            i += 1
        return C

    def extract_basis_vectors(incidence_matrix):
        """
        The name of the method describes what we want t achieve. We calculate the nullspace of the transposed identity matrix.
        :param incidence_matrix: Numpy Array
        :return: a collection of numpy arrays that form a base of transposed A
        """
        # To have the same dimension as described as in https://www7.in.tum.de/~esparza/fcbook-middle.pdf and to get the correct nullspace, we have to transpose
        A = np.transpose(incidence_matrix)
        # exp from book https://www7.in.tum.de/~esparza/fcbook-middle.pdf
        x = sympy.Matrix(A).nullspace()
        # TODO: Question here: Will x be always rational? Depends on sympy implementation. Normaly, yes, we we will have rational results
        x = np.array(x).astype(np.float64)
        return x

    A = compute_incidence_matrix(net)
    return extract_basis_vectors(A)



import os

import pm4py
from pm4py.algo.transformation import log_to_trie
from pm4py.visualization.trie import visualizer


def execute_script():
    log = pm4py.read_xes(os.path.join("..", "tests", "input_data", "running-example.xes"))
    trie = log_to_trie.algorithm.apply(log)
    gviz = visualizer.apply(trie, parameters={"format": "svg"})
    visualizer.view(gviz)


if __name__ == "__main__":
    execute_script()



import pm4py
import os
from pm4py.algo.conformance.alignments.petri_net import algorithm as alignments
from pm4py.algo.conformance.alignments.petri_net.utils import log_enrichment


def execute_script():
    log = pm4py.read_xes(os.path.join("..", "tests", "input_data", "running-example.xes"))
    filtered_log = pm4py.filter_variants_top_k(log, 1)
    net, im, fm = pm4py.discover_petri_net_inductive(filtered_log)
    aligned_traces = alignments.apply(log, net, im, fm, parameters={"ret_tuple_as_trans_desc": True})
    enriched_log = log_enrichment.apply(log, aligned_traces)
    print(enriched_log)


if __name__ == "__main__":
    execute_script()



import os

import pm4py
from pm4py.algo.discovery.inductive import algorithm as inductive_miner
import pandas as pd

LOGS_FOLDER = "../compressed_input_data"

for log_name in os.listdir(LOGS_FOLDER):
    if "xes" in log_name or "parquet" in log_name:
        log_path = os.path.join(LOGS_FOLDER, log_name)
        print("")
        print(log_path)
        if "xes" in log_name:
            log = pm4py.read_xes(log_path)
            variants = pm4py.get_variants_as_tuples(log)
            variants = {",".join(x): y for x, y in variants.items()}
            fp_log = pm4py.algo.discovery.footprints.log.variants.entire_event_log.apply(log)
        elif "parquet" in log_name:
            dataframe = pd.read_parquet(log_path)
            variants = pm4py.get_variants_as_tuples(dataframe)
            variants = {",".join(x): y for x, y in variants.items()}
            fp_log = pm4py.algo.discovery.footprints.log.variants.entire_dataframe.apply(dataframe)
        tree_im = inductive_miner.apply_tree_variants(variants, variant=inductive_miner.Variants.IM)
        tree_imf = inductive_miner.apply_tree_variants(variants, variant=inductive_miner.Variants.IMf)
        tree_imd = inductive_miner.apply_tree_variants(variants, variant=inductive_miner.Variants.IMd)
        fp_tree_im = pm4py.algo.discovery.footprints.tree.variants.bottomup.apply(tree_im)
        fp_tree_imf = pm4py.algo.discovery.footprints.tree.variants.bottomup.apply(tree_imf)
        fp_tree_imd = pm4py.algo.discovery.footprints.tree.variants.bottomup.apply(tree_imd)
        fp_conf_im = pm4py.algo.conformance.footprints.variants.log_extensive.apply(fp_log, fp_tree_im)
        fp_conf_imf = pm4py.algo.conformance.footprints.variants.log_extensive.apply(fp_log, fp_tree_imf)
        fp_conf_imd = pm4py.algo.conformance.footprints.variants.log_extensive.apply(fp_log, fp_tree_imd)
        fitness_im = pm4py.algo.conformance.footprints.util.evaluation.fp_fitness(fp_log, fp_tree_im, fp_conf_im)
        fitness_imd = pm4py.algo.conformance.footprints.util.evaluation.fp_fitness(fp_log, fp_tree_imd, fp_conf_imd)
        fitness_imf = pm4py.algo.conformance.footprints.util.evaluation.fp_fitness(fp_log, fp_tree_imf, fp_conf_imf)
        precision_im = pm4py.algo.conformance.footprints.util.evaluation.fp_precision(fp_log, fp_tree_im)
        precision_imf = pm4py.algo.conformance.footprints.util.evaluation.fp_precision(fp_log, fp_tree_imf)
        precision_imd = pm4py.algo.conformance.footprints.util.evaluation.fp_precision(fp_log, fp_tree_imd)
        print("IM fp-fitness=%.3f fp-precision=%.3f" % (fitness_im, precision_im))
        print("IMf fp-fitness=%.3f fp-precision=%.3f" % (fitness_imf, precision_imf))
        print("IMd fp-fitness=%.3f fp-precision=%.3f" % (fitness_imd, precision_imd))



import pandas as pd

from pm4py.objects.log.util import dataframe_utils
from pm4py.util import pandas_utils


def execute_script():
    # loads a dataframe. setup dates
    df = pd.read_csv("../tests/input_data/receipt.csv")
    df = dataframe_utils.convert_timestamp_columns_in_df(df)
    print(df)
    # insert the case index in the dataframe
    df = pandas_utils.insert_ev_in_tr_index(df, case_id="case:concept:name", column_name="@@index_in_trace")
    # filter all the prefixes of length 5 from the dataframe
    df = df[df["@@index_in_trace"] <= 5]
    print(df)
    # performs the automatic feature extraction
    fea_df = dataframe_utils.automatic_feature_extraction_df(df)
    print("\nfea_df =")
    print(fea_df)
    print(fea_df.columns)


if __name__ == "__main__":
    execute_script()
